/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include <linux/slab.h>
#include <linux/kdev_t.h>
#include <linux/sched.h>
#include <linux/log2.h>

#include "cow.h"
#include "debug.h"
#include "uassert.h"
#include "portability.h"

static void add_cow_metadata_page(struct cow_storage *cstor, unsigned long addr)
{
	if (!cstor->metapages)
		cstor->metapages = addr;  /* set head */
	else
		*(unsigned long *)cstor->metapage = addr;  /* set ->next */
	*(unsigned long *)addr = 0;
	cstor->metapage = addr;
	cstor->metapage_off = sizeof(unsigned long);
	cstor->num_metapages++;
}

static void free_cow_metadata_pages(struct cow_storage *cstor)
{
	unsigned long addr = cstor->metapages;

	while (addr) {
		unsigned long prev = addr;
#if defined(DEBUG)
		int i, j, cows_per_page;
		struct cow_buffer *cow = (struct cow_buffer *)(addr + sizeof(unsigned long));

		if (addr == cstor->metapage)  /* last page? */
			cows_per_page = cstor->metapage_off / sizeof(struct cow_buffer);
		else
			cows_per_page = (PAGE_SIZE - sizeof(unsigned long)) / sizeof(struct cow_buffer);

		for (i = 0; i < cows_per_page; ++i) {
			int leaked = 0;

			for (j = 0; j < PAGES_PER_COW; ++j) {
				if (cow->pages[j]) {
					leaked = 1;
					break;
				}
			}
			if (leaked) {
				stcvsm_err(LOG_COWS, "COW with leaked pages:\n"
						"\tCOW %u  state %u  refs %u  readers %u  error %d\n"
						"\ttier private %llx %llx\n"
						"\tpages %p %p %p...\n",
						(unsigned int)cow->index, (unsigned int)cow->state,
						(unsigned int)cow->refs, (unsigned int)cow->readers,
						(int)cow->error, cow->tier_private[0],
						cow->tier_private[1], cow->pages[0], cow->pages[1],
						cow->pages[2]);
			}
			cow++;
		}
#endif
		addr = *(unsigned long *)addr;
		free_page(prev);
	}
}

static struct cow_buffer *alloc_cow(struct cow_storage *cstor)
{
	struct cow_buffer *cow = (struct cow_buffer *)(cstor->metapage + cstor->metapage_off);

	cstor->metapage_off += sizeof(struct cow_buffer);
	if (cstor->metapage_off + sizeof(struct cow_buffer) > PAGE_SIZE)
		cstor->metapage_off = 0;
	return cow;
}

static void init_cow(struct cow_buffer *cow, u32 index)
{
	memset(cow, 0, sizeof(*cow));

	RB_CLEAR_NODE(&cow->node);
	INIT_LIST_HEAD(&cow->lru);
	cow->index = index;
	cow->stamp = jiffies;
}

static void recycle_cow(struct cow_buffer *cow)
{
	cow_state_trans(cow, COW_NEW);
	init_cow(cow, (u32)-1);
}

struct cow_storage *cstor_construct(sector_t sectors, dev_t orig_dev)
{
	struct cow_storage *cstor;
	unsigned int zone_bytes;
	unsigned int cows;

	cstor = kzalloc(sizeof(*cstor), GFP_KERNEL);
	if (!cstor)
		goto out1;
	cstor->root = RB_ROOT;
	INIT_LIST_HEAD(&cstor->free);
	cstor->sectors = sectors;
	cstor->sectors_per_cow = (PAGE_SIZE * PAGES_PER_COW) >> 9;
	cstor->sector_shift = ilog2(cstor->sectors_per_cow);
	if (((sectors-1) >> cstor->sector_shift) > 0xffffffff) {
		/* cow_buffer.index is 32 bit, but should be sufficient:
		 * 2**32 * 64k COWS = 256TB volumes
		 */
		stcvsm_err(LOG_COWS, "volume size too large: %llu sectors\n",
				(unsigned long long)sectors);
		goto out2;
	}
	cows = (sectors + cstor->sectors_per_cow - 1) >> cstor->sector_shift;
	if (cows < PAGE_SIZE*8) {
		zone_bytes = (cows+7)/8;
	} else {
		zone_bytes = PAGE_SIZE;
		do {
			cstor->zone_shift++;
		} while ((cows >> cstor->zone_shift) > zone_bytes*8);
		zone_bytes = (cows >> cstor->zone_shift)/8;
	}
	init_waitqueue_head(&cstor->waiters);
	cstor->cow_zones = kzalloc(zone_bytes, GFP_KERNEL);
	if (!cstor->cow_zones)
		goto out2;
	wq_init(&cstor->flush_wq, "stcvsmflush-%d:%d", MAJOR(orig_dev), MINOR(orig_dev));
	cstor->name = cstor->flush_wq.name;
	if (wq_start(&cstor->flush_wq)) {
		stcvsm_err(LOG_COWS, "%s failed to start flush wq\n", cstor->name);
		goto out3;
	}
	wq_init(&cstor->cow_wq, "stcvsm-%d:%d", MAJOR(orig_dev), MINOR(orig_dev));
	if (wq_start(&cstor->cow_wq)) {
		stcvsm_err(LOG_COWS, "%s failed to start cow wq\n", cstor->name);
		goto out4;
	}

	stcvsm_info(LOG_COWS, "%s allocated for %llu sectors; %u sectors per cow; %u cows; %u zone shift, %u zone bytes\n",
			cstor->name, (unsigned long long)sectors, cstor->sectors_per_cow, cows,
			cstor->zone_shift, zone_bytes);
	stcvsm_debug(LOG_COWS, "%lu bytes/cow\n", sizeof(struct cow_buffer));
	return cstor;

out4:
	wq_deinit(&cstor->cow_wq);
	wq_stop(&cstor->flush_wq, NULL, NULL);
out3:
	wq_deinit(&cstor->flush_wq);
	kfree(cstor->cow_zones);
out2:
	kfree(cstor);
out1:
	return NULL;
}

int cstor_add_tier(struct cow_storage *cstor, struct cow_ops *ops, ...)
{
	unsigned int depth = 0;
	struct cow_tier *tier;
	va_list args;
	int err;
	struct cow_tier **parent;

	if (!cstor->tiers) {
		parent = &cstor->tiers;
	} else {
		struct cow_tier *t = cstor->tiers;

		for (depth++; t->next; t = t->next)
			depth++;
		parent = &t->next;
	}

	if (depth >= MAX_COW_TIERS) {
		err = -E2BIG;
		goto fail;
	}

	va_start(args, ops);
	tier = ops->cow_init(cstor, args);
	va_end(args);
	if (IS_ERR(tier)) {
		err = PTR_ERR(tier);
		goto fail;
	}
	*parent = tier;
	tier->depth = depth;
	stcvsm_info(LOG_COWS, "%s Added tier %s at depth %u\n", cstor->name, ops->cow_name, depth);
	return 0;

fail:
	stcvsm_err(LOG_COWS, "%s Failed to add tier %s (%d)\n", cstor->name, ops->cow_name, err);
	return err;
}

static void cstor_free(void *data)
{
	struct cow_storage *cstor = (struct cow_storage *)data;
	struct cow_tier *tier;

	stcvsm_info(LOG_COWS, "%s\n", cstor->name);

	tier = cstor->tiers;
	while (tier) {
		struct cow_tier *next = tier->next;

		tier->ops->cow_deinit(tier);
		tier = next;
	}

	wq_stop(&cstor->flush_wq, NULL, NULL);
	wq_deinit(&cstor->flush_wq);
	wq_deinit(&cstor->cow_wq);
	free_cow_metadata_pages(cstor);
	kfree(cstor->cow_zones);
	kfree(cstor);
}

void cstor_shutdown(struct cow_storage *cstor)
{
	struct cow_tier *tier;

	stcvsm_notice(LOG_COWS, "%s shutting down\n", cstor->name);
	cstor_lock_irq(cstor);
	cstor->shutdown = 1;
	cstor_unlock_irq(cstor);

	/* Shut down any IO occurring internally to the cstor. */
	for (tier = cstor->tiers; tier; tier = tier->next) {
		if (tier->ops->cow_shutdown)
			tier->ops->cow_shutdown(tier);
	}
}

void cstor_destruct(struct cow_storage *cstor)
{
	cstor_shutdown(cstor);
	/* Probably called from cow_wq (certainly not flush_wq) so stop and unwind myself first */
	wq_stop(&cstor->cow_wq, cstor_free, cstor);
}

static int cow_insert(struct rb_root *root, struct cow_buffer *cow)
{
	struct rb_node **new = &(root->rb_node), *parent = NULL;

	/* Figure out where to put new node */
	while (*new) {
		struct cow_buffer *this = container_of(*new, struct cow_buffer, node);

		parent = *new;
		if (cow->index < this->index) {
			new = &((*new)->rb_left);
		} else if (cow->index > this->index) {
			new = &((*new)->rb_right);
		} else {
			UASSERT(0, "Failure to insert a COW");
			return -1;
		}
	}

	/* Add new node and rebalance tree. */
	rb_link_node(&cow->node, parent, new);
	rb_insert_color(&cow->node, root);

	return 0;
}

static struct cow_buffer *cow_search(struct rb_root *root, u32 index)
{
	struct rb_node *node = root->rb_node;

	while (node) {
		struct cow_buffer *cow = container_of(node, struct cow_buffer, node);

		if (index < cow->index)
			node = node->rb_left;
		else if (index > cow->index)
			node = node->rb_right;
		else
			return cow;
	}
	return NULL;
}

struct cow_buffer *cow_get(struct cow_storage *cstor, u32 index)
{
	struct cow_buffer *cow = NULL;
	unsigned long zone = index >> cstor->zone_shift;
	unsigned long page;

	if ((index << cstor->sector_shift) >= cstor->sectors)
		return NULL;

	cstor_lock_irq(cstor);
do_search:
	if (test_bit(zone, cstor->cow_zones))
		cow = cow_search(&cstor->root, index);
	if (!cow) {
		/* Prefer getting off free list */
		if (!list_empty(&cstor->free)) {
			cow = list_entry(cstor->free.next, struct cow_buffer, lru);
			list_del(&cow->lru);
		} else {
			if (!cstor->metapage_off) {
				cstor_unlock_irq(cstor);
				if (ERROR_INJECT(ERRINJ_ALLOC))
					page = 0;
				else
					page = __get_free_page(GFP_KERNEL);
				if (!page) {
					stcvsm_err(LOG_COWS, "%s failed to allocate cow metadata page\n",
							cstor->name);
					return NULL;
				}
				stcvsm_debug(LOG_COWS, "%s allocated cow metadata page %p\n",
						cstor->name, (void *)page);

				cstor_lock_irq(cstor);
				if (!cstor->metapage_off) {
					add_cow_metadata_page(cstor, page);
				} else {
					cstor_unlock_irq(cstor);
					free_page(page);
					cstor_lock_irq(cstor);
				}
				goto do_search;
			}
			cow = alloc_cow(cstor);
		}

		init_cow(cow, index);
		set_bit(zone, cstor->cow_zones);
		cow_insert(&cstor->root, cow);
		cstor->num_cows++;
	}
	cow->refs++;
	cstor_unlock_irq(cstor);

	stcvsm_trace(LOG_COWS, "%s cow %u state %u refs %u\n", cstor->name, (unsigned int)index,
			cow->state, cow->refs);

	return cow;
}

/*
 * perform COW state transition
 *  generate traces from that
 */
void cow_state_trans(struct cow_buffer *cow, int state)
{
	unsigned long now = jiffies;
	switch (state) {
	case COW_NEW:
		trace_cow_state_trans_new(cow, NULL, cow->state, state, now - cow->stamp);
		break;
	case COW_FILLING:
		trace_cow_state_trans_filling(cow, NULL, cow->state, state, now - cow->stamp);
		break;
	case COW_PAGEDIN:
		trace_cow_state_trans_pagedin(cow, NULL, cow->state, state, now - cow->stamp);
		break;
	case COW_PAGEDOUT:
		trace_cow_state_trans_pagedout(cow, NULL, cow->state, state, now - cow->stamp);
		break;
	case COW_PAGINGIN:
		trace_cow_state_trans_pagingin(cow, NULL, cow->state, state, now - cow->stamp);
		break;
	case COW_ERROR:
		trace_cow_state_trans_error(cow, NULL, cow->state, state, now - cow->stamp);
		break;
	case COW_BARRIER:
		trace_cow_state_trans_barrier(cow, NULL, cow->state, state, now - cow->stamp);
		break;
	default:
		trace_cow_state_trans(cow, NULL, cow->state, state, now - cow->stamp);
		break;
	}
	cow->stamp = now;
	cow->state = state;
}

/*
 * Allocates pages to back the COW.
 */
int cow_alloc_pages(struct cow_storage *cstor, struct page **pages)
{
	int i;

	UASSERT(!pages[0], "Pages already allocated");

	for (i = 0; i < PAGES_PER_COW; ++i) {
		if (ERROR_INJECT(ERRINJ_ALLOC))
			pages[i] = NULL;
		else
			/* Specifically accept highmem (if any) to lessen normal memory pressure */
			pages[i] = alloc_page(GFP_NOFS | __GFP_HIGHMEM);
		if (!pages[i]) {
			while (i > 0) {
				__free_page(pages[--i]);
				pages[i] = NULL;
			}
			return -ENOMEM;
		}
	}

	atomic_inc(&cstor->nr_active_cows);
	return 0;
}

void cow_free_pages(struct cow_storage *cstor, struct page **pages)
{
	int i;

	for (i = 0; i < PAGES_PER_COW; ++i) {
		UASSERT(pages[i], "Page not available to free");
		__free_page(pages[i]);
		pages[i] = NULL;
	}
	atomic_dec(&cstor->nr_active_cows);
}

int cow_start_fill(struct cow_storage *cstor, struct cow_buffer *cow)
{
	int r;

	cstor_lock_irq(cstor);

	if (cow->state == COW_FILLING) {
		r = -EBUSY;
	} else if (cow->state != COW_NEW && cow->state != COW_BARRIER) {
		r = 1;
	} else if (cstor->shutdown) {
		r = -EIO;
	} else {
		cow_state_trans(cow, COW_FILLING);
		cstor->stores++;
		r = 0;
	}
	cstor_unlock_irq(cstor);

	if (r == 0) {
		struct cow_tier *tier;

		stcvsm_debug(LOG_COWS, "%s cow %u refs %u\n", cstor->name, (unsigned int)cow->index,
				cow->refs);

		for (tier = cstor->tiers; tier; tier = tier->next) {
			if (tier->ops->cow_wait)
				tier->ops->cow_wait(tier);
		}

		r = cow_alloc_pages(cstor, cow->pages);
		if (r) {
			cstor_lock_irq(cstor);
			cow_state_trans(cow, COW_ERROR);
			cow->error = r;
			cstor->errors++;
			cstor_unlock_irq(cstor);

			wake_up_all(&cstor->waiters);
		}
	}

	return r;
}

void cow_reader_unlock(struct cow_storage *cstor, struct cow_buffer **cows, int n)
{
	int i;
	int wake = 0;
	int cow_idx = 0;

	for (i = 0; i < n; i++) {
		struct cow_buffer *cow = cows[i];

		cstor_lock_irqsave(cstor);
		UASSERT(cow->readers > 0, "No Readers to unlock");
		if (--cow->readers == 0) {
			wake = 1;
			cow_idx = i;
		}
		cstor_unlock_irqrestore(cstor);
	}

	if (wake) {
		stcvsm_trace(LOG_COWS, "%s wake %u refs %u\n", cstor->name,
				(unsigned int)cows[cow_idx]->index, cows[cow_idx]->refs);
		wake_up_all(&cstor->waiters);
	}
}

static inline int cow_fill_done(struct cow_buffer *cow)
{
	return cow->state != COW_FILLING && cow->readers == 0;
}

int cow_wait_fill(struct cow_storage *cstor, struct cow_buffer **cows, int n)
{
	int i, r = 0;

	for (i = 0; i < n; i++) {
		struct cow_buffer *cow = cows[i];

		if (cow) {
			cstor_lock_irq(cstor);
			if (!cow_fill_done(cow)) {
#if STCVSM_LOG_LEVEL >= LOGLEVEL_DEBUG
				u64 start = get_jiffies_64();
#endif
				stcvsm_debug(LOG_COWS, "%s waiting for cow %u to fill from %u\n",
						cstor->name, (unsigned int)cow->index, cow->state);
				stcvsm_wait_event_locked_irq(cstor->waiters, cow_fill_done(cow));
				stcvsm_debug(LOG_COWS, "%s waited %u ms for cow %u to fill: %d, %d\n",
						cstor->name,
						jiffies_to_msecs(get_jiffies_64() - start),
						(unsigned int)cow->index,
						cow->state,
						cow->state == COW_ERROR ? cow->error : 0);
			}

			if (cow->state == COW_ERROR) {
				r = cow->error;
				/* Even if the final state is an error, must wait on all before
				 * allowing writes.  Some may still succeed.
				 */
			}
			cstor_unlock_irq(cstor);
		}
	}
	return r;
}

static inline int cow_pagein_done(struct cow_buffer *cow)
{
	return cow->state != COW_FILLING && cow->state != COW_PAGINGIN;
}

int cow_wait_pagein(struct cow_storage *cstor, struct cow_buffer **cows, int n)
{
	int i, r = 0;

	/* Stop waiting at the first failure */
	for (i = 0; i < n && r == 0; i++) {
		struct cow_buffer *cow = cows[i];

		if (cow) {
			cstor_lock_irq(cstor);
			if (!cow_pagein_done(cow)) {
#if STCVSM_LOG_LEVEL >= LOGLEVEL_DEBUG
				u64 start = get_jiffies_64();
#endif
				stcvsm_debug(LOG_COWS, "%s waiting for cow %u to pagein from %u\n",
						cstor->name, (unsigned int)cow->index, cow->state);
				stcvsm_wait_event_locked_irq(cstor->waiters, cow_pagein_done(cow));
				stcvsm_debug(LOG_COWS, "%s waited %u ms for cow %u to pagein: %d, %d\n",
						cstor->name,
						jiffies_to_msecs(get_jiffies_64() - start),
						(unsigned int)cow->index,
						cow->state,
						cow->state == COW_ERROR ? cow->error : 0);
			}

			if (cow->state == COW_ERROR)
				r = cow->error;
			cstor_unlock_irq(cstor);
		}
	}
	return r;
}

void cow_put(struct cow_storage *cstor, struct cow_buffer *cow)
{
	int error = 0;
	int wake = 0;

	cstor_lock_irqsave(cstor);
	UASSERT(cow->refs > 0, "No REFs to PUT");
	if (--cow->refs == 0) {
		UASSERT(cow->readers == 0, "Readers should be zero");  /* cow_reader_unlock should be nested */
		UASSERT(cow->state != COW_PAGINGIN, "COW state should not be PAGINGIN");
		stcvsm_trace(LOG_COWS, "%s cow %u state %u\n", cstor->name, (unsigned int)cow->index,
				cow->state);
		if (cow->state == COW_BARRIER || cow->state == COW_NEW) {
			/* No writer has raced and claimed these?  Put back on free list. */
			UASSERT(!cow->pages[0], "No page");
			rb_erase(&cow->node, &cstor->root);
			recycle_cow(cow);
			list_add_tail(&cow->lru, &cstor->free);
			cstor->num_cows--;
		} else if (cow->state == COW_PAGEDIN) {
			wake = 1;
			if (cstor->shutdown) {
				error = 1;
				cow->error = ENODEV;
				cow_state_trans(cow, COW_ERROR);
			} else {
				cow_state_trans(cow, COW_PAGEDOUT);
				cstor->tiers->ops->cow_push(cstor->tiers, cow);
			}
		}
	}
	cstor_unlock_irqrestore(cstor);

	if (error)
		cow_free_pages(cstor, cow->pages);
	if (wake)
		wake_up_all(&cstor->waiters);
}

void cow_finish_fill(struct cow_storage *cstor, struct cow_buffer **cows, int n, int error)
{
	int i;

	for (i = 0; i < n; ++i) {
		struct cow_buffer *cow = cows[i];

		UASSERT(cow->state == COW_FILLING, "COW state not filling");
		cstor_lock_irqsave(cstor);
		if (error) {
			cow->error = error;
			cow_state_trans(cow, COW_ERROR);
		} else {
			cow_state_trans(cow, COW_PAGEDIN);
		}
		cstor_unlock_irqrestore(cstor);

		stcvsm_debug(LOG_COWS, "%s filled cow %u error %d refs %u\n", cstor->name,
				(unsigned int)cow->index, error, cow->refs);

		if (error)
			cow_free_pages(cstor, cow->pages);
	}
	wake_up_all(&cstor->waiters);
}

static void pagedin_callback(struct cow_tier *tier, struct cow_buffer *cow, int error)
{
	struct cow_storage *cstor = tier->cstor;

	cstor_lock_irqsave(cstor);
	if (error) {
		cow->error = error;
		cow_state_trans(cow, COW_ERROR);
	} else {
		cow_state_trans(cow, COW_PAGEDIN);
	}
	cstor_unlock_irqrestore(cstor);

	stcvsm_debug(LOG_COWS, "%s cow %u state %u refs %u\n", cstor->name, (unsigned int)cow->index,
			cow->state, cow->refs);
	cow_put(cstor, cow);

	wake_up_all(&cstor->waiters);
}

int cow_start_pagein(struct cow_storage *cstor, struct cow_buffer *cow)
{
	int r = 0;
	int pagein = 0;

	cstor_lock_irq(cstor);

	/* Readers to new indicies must block racing writers */
	if (cow->state == COW_NEW) {
		UASSERT(cow->readers == 0, "COW Readers should be 0 at this time");
		cow_state_trans(cow, COW_BARRIER);
	}
	if (cow->state == COW_BARRIER) {
		cow->readers++;
		r = 1;
	} else if (cow->state == COW_PAGEDIN || cow->state == COW_PAGINGIN ||
			cow->state == COW_FILLING) {
		cstor->loads++;
		cstor->hits++;
	} else if (cstor->shutdown) {
		r = -EIO;
	} else if (cow->state == COW_PAGEDOUT) {
		cow_state_trans(cow, COW_PAGINGIN);
		cstor->loads++;
		cow->refs++;  /* Hold until pagedin_callback */
		pagein = 1;
	} else if (cow->state == COW_ERROR) {
		r = cow->error;
		cstor->errors++;
	} else {
		UASSERT(0, "Bad COW state");
		r = -EINVAL;
	}

	cstor_unlock_irq(cstor);

	if (pagein) {
		stcvsm_debug(LOG_COWS, "%s cow %u\n", cstor->name, (unsigned int)cow->index);
		cstor->tiers->ops->cow_pull(cstor->tiers, cow, pagedin_callback);
	} else if (r == 1) {
		stcvsm_trace(LOG_COWS, "%s barrier %u\n", cstor->name, (unsigned int)cow->index);
	}
	return r;
}

void cow_log_stats(struct cow_storage *cstor, struct seq_file *s)
{
	struct cow_tier *tier;

	seq_printf(s, " COW manager:\t%lu stores  %lu loads  %lu hits  %lu errors  %lu tot %u act cows  %lu kb metadata\n",
			cstor->stores, cstor->loads, cstor->hits, cstor->errors, cstor->num_cows,
			atomic_read(&cstor->nr_active_cows), cstor->num_metapages*(PAGE_SIZE/1024));
	for (tier = cstor->tiers; tier; tier = tier->next) {
		seq_printf(s, " COW L%d (%s):\t%lu stores  %lu loads  %lu hits  %d errors  %lu/%lu kb  ",
				tier->depth, tier->ops->cow_name, tier->stores, tier->loads,
				tier->hits, atomic_read(&tier->errors), tier->kb, tier->max_kb);
		if (tier->ops->cow_log)
			tier->ops->cow_log(tier, s);
		seq_puts(s, "\n");
	}
}

int cow_pack_bios(struct cow_storage *cstor, struct cow_buffer *cow, unsigned long cow_index,
		struct block_device *bdev, struct bio **bios)
{
	unsigned int num_pages = PAGES_PER_COW;
	unsigned int size = PAGES_PER_COW * PAGE_SIZE;
	unsigned int tail_size = PAGE_SIZE;
	unsigned int nr_accepted = 0;
	int num_bios = 0;
	sector_t start_sector = (sector_t)cow_index * (size>>9);
	struct bio *bio = NULL;
	struct bio *prev = NULL;
	struct bio *head = NULL;
	int retval;

	/* Truncate size due to end of disk? */
	if (cow_to_sector(cstor, cow) >= cstor->sectors)
		return -EIO;
	if (cstor->sectors - cow_to_sector(cstor, cow) < cstor->sectors_per_cow) {
		size = (cstor->sectors - cow_to_sector(cstor, cow)) << 9;
		if (offset_in_page(size))
			tail_size = offset_in_page(size);
		num_pages = (size + PAGE_SIZE - 1) >> PAGE_SHIFT;
		stcvsm_info(LOG_COWS, "truncating sector %llu to %u bytes for cow %u\n",
				(unsigned long long)start_sector, size, cow->index);
	}

	/* Add pages to a bio until it hits the (dynamic) limit */
	while (1) {
		int accepted, empty = 0;

		if (!bio) {
			if (ERROR_INJECT(ERRINJ_ALLOC))
				bio = NULL;
			else
				bio = bio_alloc(GFP_NOIO, 1);
			if (!bio) {
				stcvsm_err(LOG_COWS, "bio_alloc failed for cow %u\n",
						(unsigned int)cow->index);
				retval = -ENOMEM;
				goto fail;
			}
			num_bios++;
			bio->bi_next = NULL;
			if (!head)
				head = bio;
			else
				prev->bi_next = bio;
			prev = bio;
			stc_bio_assign_disk(bio, bdev);
			stc_bio_set_op(bio, READ);
			bio_sector(bio) = start_sector + nr_accepted*(PAGE_SIZE>>9);
			empty = 1;
		}
		accepted = bio_add_page(bio, cow->pages[nr_accepted],
				nr_accepted+1 == num_pages ? tail_size : PAGE_SIZE, 0);
		if (accepted) {
			nr_accepted++;
		} else if (empty) {
			stcvsm_err(LOG_COWS, "bio_add_page failed for cow %u page %u/%u %p size %u\n",
					(unsigned int)cow->index, nr_accepted, num_pages,
					cow->pages[nr_accepted], size);
			retval = -EIO;
			goto fail;
		}
		if (!accepted || nr_accepted == num_pages) {
			bio = NULL;
			if (nr_accepted == num_pages)
				break;
		}
	}

	*bios = head;
	return num_bios;

fail:
	for (bio = head; bio; bio = head) {
		head = bio->bi_next;
		bio->bi_next = NULL;
		bio_put(bio);
	}
	*bios = NULL;
	return retval;
}
