/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#ifndef STCVSM_COW_H
#define STCVSM_COW_H

#include <linux/bio.h>
#include <linux/rbtree.h>
#include <linux/wait.h>
#include <linux/seq_file.h>
#include <stdarg.h>

#include "workqueue.h"

/** Copy-on-write tiered storage API.
 *
 *
 * VOLUME WRITE
 *
 * - cow_start_fill() should be matched by cow_finish_fill().
 * - cow_wait_fill() waits for data to be filled into the cow, but also is a writer lock, which
 *   waits for all snapshot readers to finish.
 *
 * volume request_fn		volume read endio		workqueue
 * -----------------		-----------------		---------
 * cow = cow_get(index);
 * cow_start_fill(cow);
 * submit_bio() // fill cow
 * wq_insert_work()
 *				cow_finish_fill(cow);		cow_wait_fill(cow);
 *								cow_put(cow)
 *								orig_reqfn();
 *
 * SNAPSHOT READ
 *
 * - cow_start_pagein() claims reader locks as necessary, which must be matched by a call to
 *   cow_reader_unlock().
 * - The "volume read endio" swim lane is omitted entirely if no read locks were taken (that is, if
 *   nothing is read from the volume).
 *
 *
 * snapshot request_fn		volume read endio		workqueue
 * -------------------		-----------------		---------
 * cow = cow_get(index);
 * cow_start_pagein(cow);
 * bio_clone() // read volume
 *				cow_reader_unlock(cow);
 *				wq_insert_work()
 *								cow_wait_pagein(cow);
 *								// overlay data from cows
 *								cow_put(cow);
 */

#define COW_NEW       0  /* newly allocated; progress to FILLING or BARRIER via cow_start_fill */
#define COW_FILLING   1  /* busy filling; progress to PAGEDIN or ERROR via cow_finish_fill */
#define COW_PAGEDIN   2  /* filled and in memory; may return to PAGEDOUT via cow_put */
#define COW_PAGEDOUT  3  /* filled but not in memory; progress to PAGEDIN via cow_pagein */
#define COW_PAGINGIN  4  /* progressing PAGEDOUT -> (PAGEDIN, ERROR) */
#define COW_ERROR     5  /* IO error; terminal state */
#define COW_BARRIER   6  /* prevents creation of permanent cow here until read(s) complete */

/* Pages per cow affects overhead and performance:
 *	pages	kb	cow_buffer	MB overhead
 *			size		per 1GB COWs
 *	1	4	64		14
 *	2	8	72		10
 *	4	16	88		6
 *	8	32	120		3.7
 *	16	64	184		2.9
 *	32	128	312		2.5
 */
#define PAGES_PER_COW 16

/* Maximum number of tiers that can be chained.  Only reason to limit this is keep cow_buffer small.
 */
#define MAX_COW_TIERS 2

struct cow_tier;
struct cow_storage;

/** Represents a copy of one or more sectors (a "cow").  Every cow known to the system, even those
 * pushed into the tiers, are represented by a cow_buffer so keep this small.
 *
 * Copies of sectors are packed onto page(s).  All cow operations affect the entire cow, not
 * individual sectors.
 */
struct cow_buffer {
	struct rb_node node;	/* tracking by the red-black tree */
	struct list_head lru;	/* tracking on lists in tiers */
	u32 index;		/* index used as key for rbtree */
	unsigned short state;	/* state of the cow */
	unsigned short refs;	/* # of references */
	unsigned short readers; /* # of readers remaining, which started as BARRIER */
	short error;		/* cause of COW_ERROR state */
	u64 tier_private[MAX_COW_TIERS];
	struct page *pages[PAGES_PER_COW];
	unsigned long stamp;	/* date stamp used for timing metrics */
};

typedef void (*cow_callback)(struct cow_tier *tier, struct cow_buffer *cow, int error);

/** Function pointer table to implement a cow_tier.
 */
struct cow_ops {
	/** Initialize a new cow_tier struct */
	struct cow_tier * (*cow_init)(struct cow_storage *, va_list);
	void (*cow_push)(struct cow_tier *tier, struct cow_buffer *cow);
	void (*cow_pull)(struct cow_tier *tier, struct cow_buffer *cow, cow_callback);
	void (*cow_log)(struct cow_tier*, struct seq_file *s);
	/** Blocks until sufficient in-flight IO catches up */
	void (*cow_wait)(struct cow_tier *);
	/** Shuts down before deinit. */
	void (*cow_shutdown)(struct cow_tier *);
	void (*cow_deinit)(struct cow_tier *);
	const char *cow_name;
};

/** A tier in the storage hierarchy of cows.  The cow store, and the various tiers, can push down /
 * pull up the stack.  A cow is owned by at most a single tier.  Tiers can be stacked arbitrarily,
 * up to MAX_COW_TIERS.
 *
 * A tier knows its depth, and can have a single entry on the cow to help track metadata.  While a
 * tier owns a cow, it is free to use cow->lru and the pages array.
 *
 * When pulling, it is the pullee who must allocate (or surrender existing) pages.  Initiating a
 * pull can block due to memory pressure.
 *
 * Pushing does not block and cannot fail, because it is just passing ownership of pages down.  This
 * implies that committing to the tier's permanent storage may be asynchronous.
 *
 * Pushing implies relinquishing ownership.  Therefore pulling can only come from a single tier,
 * and there is no confusion about who controls the pages array at any time.
 *
 * cow->state and cow->refs should almost never referenced by the tiers.  For a cow to get pushed
 * into the tiers, it is PAGEDOUT and unreferenced.  The only exception is if a tier must drop a
 * cow due to running out of space.  For this it can set the cow into an error state.
 *
 * A cow will only experience a single push or pull at a time, due to the state machine in cow.c.
 */
struct cow_tier {
	struct cow_ops *ops;
	struct cow_tier *next;
	struct cow_storage *cstor;
	unsigned int depth;
	unsigned long stores;
	unsigned long loads;
	unsigned long hits;
	atomic_t errors;
	unsigned long kb;
	unsigned long max_kb;
};

/** The public object representing the cow storage hierarchy.
 */
struct cow_storage {
	struct rb_root root;		/* rbtree of all cows */

	unsigned long metapages;	/* head of list of pages containing all cow_buffers */
	unsigned long metapage;		/* current page */
	unsigned int metapage_off;	/* next free offset or 0 */
	unsigned int num_metapages;
	struct list_head free;		/* free cows */

	struct cow_tier *tiers;		/* list of tiers */

	unsigned long stores;
	unsigned long loads;
	unsigned long hits;
	unsigned long errors;
	unsigned long num_cows;
	atomic_t nr_active_cows;

	unsigned int zone_shift;
	/** Optional bitmap of "zones" of cows; 0 implies no cows in this range.  Optimization to
	 * avoid rbtree */
	unsigned long *cow_zones;

	sector_t sectors;
	unsigned int sector_shift;
	unsigned int sectors_per_cow;

	wait_queue_head_t waiters;  /* waiters for cow status change */
	struct wq cow_wq;  /* copy-on-write workqueue */
	struct wq flush_wq;  /* flusher */
	const char *name;

	unsigned long lflags;
	int shutdown;
};

static inline u32 sector_to_index(struct cow_storage *cstor, sector_t sector)
{
	return (u32)(sector >> cstor->sector_shift);
}
static inline sector_t cow_to_sector(struct cow_storage *cstor, struct cow_buffer *cow)
{
	return ((sector_t)cow->index) << cstor->sector_shift;
}

/** Construct a cow_storage.  Must also call cow_add_tier at least once.
 * @param sectors  Number of sectors cstor must track
 * @param orig_dev Major/minor of device to copy-on-write
 */
struct cow_storage *cstor_construct(sector_t sectors, dev_t orig_dev);

/** Additional args are passed through to cow_init.
 */
int cstor_add_tier(struct cow_storage *cstor, struct cow_ops *ops, ...);

/** Optionally idle activity and close devices before destruction.
 */
void cstor_shutdown(struct cow_storage *cstor);

/** Shuts down and destroys a cow storage.  All cows are destroyed.
 */
void cstor_destruct(struct cow_storage *cstor);

/** Return a cow_buffer for the index.  This does not ensure that sufficient backing pages exist
 * (see cow_start_fill or cow_start_pagein).
 *
 * @return The cow.  NULL indicates ENOMEM.
 */
struct cow_buffer *cow_get(struct cow_storage *cstor, u32 cow_index);

/** Releases a reference to the cow.  Once the reference reaches zero, the cow progresses to
 * PAGEDOUT.
 */
void cow_put(struct cow_storage *cstor, struct cow_buffer *cow);

/** Call when you intend to fill a cow.  If the cow is in the NEW state, ensures a backing
 * page is allocated and returns 0.  Once the cow is filled or failed, call cow_finish_fill to
 * advance the state and wake waiters.  If the cow is already complete, returns 1.  On error,
 * returns <0 (in particular, a full cstor is represented by -ENOSPC and a filling cow is -EBUSY).
 *
 * @return 0 on success (proceed with fill), 1 if already filled, <0 on error
 */
int cow_start_fill(struct cow_storage *cstor, struct cow_buffer *cow);

/** Waits for the cows to be filled, and for any/all readers to the original volume to be done.
 * When this returns, writes can proceed.
 */
int cow_wait_fill(struct cow_storage *cstor, struct cow_buffer **cows, int n);

/** Signals that the cow's buffer has been filled (or failed).  Cow must be in FILLING state
 * due to cow_start_fill.  Progresses to PAGEDIN or ERROR state and wakes waiters.
 *
 * May be called from interrupt context.
 *
 * @param error 0 iff successfully filled.
 */
void cow_finish_fill(struct cow_storage *cstor, struct cow_buffer **cows, int n, int error);

/** For cows that have started or finished filling (FILLING, PAGEDIN, PAGEDOUT, PAGINGIN), begins
 * moving to PAGEDIN and returns 0.  For cows that have never been filled, moves to BARRIER and
 * returns 1.  Must match with call to cow_reader_unlock.
 *
 * @return >=0 on success (proceed to cow_wait_pagein), <0 on error
 */
int cow_start_pagein(struct cow_storage *cstor, struct cow_buffer *cow);

/** Releases reader locks obtained from cow_start_pagein.
 *
 * May be called from interrupt context.
 */
void cow_reader_unlock(struct cow_storage *cstor, struct cow_buffer **cows, int n);

/** Waits for the cow(s) to progress out of the FILLING or PAGINGIN state.  This is only safe to do
 * after obtaining a reference (see cow_start_pagein).
 *
 * @return 0 if none are FILLING or PAGINGIN (implying PAGEDIN or ERROR); !0 if any are ERROR
 */
int cow_wait_pagein(struct cow_storage *cstor, struct cow_buffer **cows, int n);

void cow_log_stats(struct cow_storage *cstor, struct seq_file *s);

/** Utility function to pack the pages of a cow onto bio(s) and return them.  It is the caller's
 * responsibility to iterate through the returned list and submit_bio() or such.
 *
 * @return On success, returns the number of bios (>0) returned.  On failure, returns -error.
 */
int cow_pack_bios(struct cow_storage *cstor, struct cow_buffer *cow, unsigned long cow_index,
		struct block_device *bdev, struct bio **bios);

/** Handles the state transition to the new state
 * Allows for tracing and performace metrics
 */
void cow_state_trans(struct cow_buffer *cow, int state);
int cow_alloc_pages(struct cow_storage *cstor, struct page **pages);
void cow_free_pages(struct cow_storage *cstor, struct page **pages);

static inline void cstor_lock_irq(struct cow_storage *cstor)
{
	spin_lock_irq(&cstor->waiters.lock);
}
static inline void cstor_unlock_irq(struct cow_storage *cstor)
{
	spin_unlock_irq(&cstor->waiters.lock);
}
static inline void cstor_lock_irqsave(struct cow_storage *cstor)
{
	spin_lock_irqsave(&cstor->waiters.lock, cstor->lflags);
}
static inline void cstor_unlock_irqrestore(struct cow_storage *cstor)
{
	spin_unlock_irqrestore(&cstor->waiters.lock, cstor->lflags);
}

#include "cow_mem.h"
#include "cow_bdev.h"

#ifdef HAVE_WAITQUEUE_ENTRY
#define cow_wait_list entry
#define cow_add_wait_tail __add_wait_queue_entry_tail
#else // HAVE_WAITQUEUE_ENTRY
#define cow_wait_list task_list
#define cow_add_wait_tail __add_wait_queue_tail
#endif // HAVE_WAITQUEUE_ENTRY

#define __stcvsm_wait_event_locked_irq(wq, condition, lock)		\
do {									\
	DEFINE_WAIT(__wait);						\
									\
	do {								\
		if (likely(list_empty(&__wait.cow_wait_list)))		\
			cow_add_wait_tail(&(wq), &__wait);		\
		set_current_state(TASK_UNINTERRUPTIBLE);		\
		spin_unlock_irq(&lock);					\
		schedule();						\
		spin_lock_irq(&lock);					\
	} while (!(condition));						\
	__remove_wait_queue(&(wq), &__wait);				\
	__set_current_state(TASK_RUNNING);				\
} while (0)

#define stcvsm_wait_event_locked_irq(wq, condition)			\
do {									\
	if (condition)							\
		break;							\
	__stcvsm_wait_event_locked_irq(wq, condition, (wq).lock);	\
} while (0)

#endif // STCVSM_COW_H
