/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

/* Memory Based storage for COW pages */

#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/delay.h>

#include "cow.h"
#include "debug.h"
#include "portability.h"

struct cow_ops cow_mem_ops;
struct cow_mem_tier {
	struct cow_tier;

	struct list_head inactive;	/* cows held by this tier */
	unsigned long num_pages;	/* pages in inactive list */
	unsigned long max_pages;	/* tuneable: hard upper limit of allocated pages */
};

static struct cow_tier *cow_mem_init(struct cow_storage *cstor, va_list args)
{
	struct cow_mem_tier *mtier;
	unsigned long max_pages = 64*1024*1024/PAGE_SIZE;
	unsigned long min_pages =  1*1024*1024/PAGE_SIZE;
	unsigned long physpages = node_present_pages(first_online_node);

	mtier = kzalloc(sizeof(*mtier), GFP_KERNEL);
	if (!mtier)
		return ERR_PTR(-ENOMEM);

	/* TODO  Thoughts on dynamically calculating reasonable sizes for memory tier:
	 *   - scale off size of cstor and/or next tier
	 *   - scale off size of physical memory
	 *   - proportion total size among snapshots based on (size, write speed)
	 *   - if no next tier, be bigger
	 *   - scaling factor configurable via /proc
	 *   - beware memory (un)hotplugging
	 */
	mtier->max_pages = cstor->sectors >> (cstor->sector_shift + 9);
	if (mtier->max_pages > physpages>>4)
		mtier->max_pages = physpages>>4;
	if (mtier->max_pages > max_pages)
		mtier->max_pages = max_pages;
	else if (mtier->max_pages < min_pages)
		mtier->max_pages = min_pages;

	INIT_LIST_HEAD(&mtier->inactive);

	mtier->ops = &cow_mem_ops;
	mtier->cstor = cstor;
	mtier->max_kb = mtier->max_pages * PAGE_SIZE / 1024;
	stcvsm_info(LOG_COWS, "%lu kb\n", mtier->max_kb);

	return (struct cow_tier *)mtier;
}

static inline void inc_cow(struct cow_mem_tier *mtier)
{
	mtier->num_pages += PAGES_PER_COW;
	mtier->kb = mtier->num_pages * PAGE_SIZE / 1024;
}

static inline void dec_cow(struct cow_mem_tier *mtier)
{
	mtier->num_pages -= PAGES_PER_COW;
	mtier->kb = mtier->num_pages * PAGE_SIZE / 1024;
}

static void cow_mem_push(struct cow_tier *tier, struct cow_buffer *cow)
{
	struct cow_mem_tier *mtier = (struct cow_mem_tier *)tier;
	struct list_head *inactive = &mtier->inactive;

	mtier->stores++;
	/* Use this tier's void * merely as a flag to indicate the cow is in the list */
	cow->tier_private[mtier->depth] = 1;
	list_add_tail(&cow->lru, inactive);
	inc_cow(mtier);

	while (mtier->num_pages > mtier->max_pages) {
		cow = list_entry(inactive->next, struct cow_buffer, lru);
		list_del_init(&cow->lru);
		cow->tier_private[mtier->depth] = 0;  /* No longer owned by this tier */
		dec_cow(mtier);

		if (mtier->next) {
			stcvsm_trace(LOG_COWS, "pushing cow %u to %s\n", (unsigned int)cow->index,
					mtier->next->ops->cow_name);
			mtier->next->ops->cow_push(mtier->next, cow);
			cow = NULL;
		} else {
			stcvsm_trace(LOG_COWS, "dropping cow %u\n", (unsigned int)cow->index);
			cow_state_trans(cow, COW_ERROR);
			cow->error = -ENOSPC;
		}

		if (cow) {
			atomic_inc(&mtier->errors);
			cow_free_pages(mtier->cstor, cow->pages);
		}
	}
}

static void cow_mem_pull(struct cow_tier *tier, struct cow_buffer *cow, cow_callback cb)
{
	struct cow_mem_tier *mtier = (struct cow_mem_tier *)tier;
	int found = 0;

	cstor_lock_irq(mtier->cstor);
	mtier->loads++;
	if (cow->tier_private[mtier->depth]) {
		mtier->hits++;
		cow->tier_private[mtier->depth] = 0;  /* No longer owned by this tier */
		list_del_init(&cow->lru);
		dec_cow(mtier);
		found = 1;
	}
	cstor_unlock_irq(mtier->cstor);

	if (found)
		cb(tier, cow, 0);
	else if (mtier->next)
		mtier->next->ops->cow_pull(mtier->next, cow, cb);
	else
		cb(tier, cow, -ENOENT);
}

static void cow_mem_deinit(struct cow_tier *tier)
{
	struct cow_mem_tier *mtier = (struct cow_mem_tier *)tier;
	struct list_head *pos;

	list_for_each(pos, &mtier->inactive) {
		struct cow_buffer *cow = list_entry(pos, struct cow_buffer, lru);

		cow_free_pages(mtier->cstor, cow->pages);
	}

	kfree(mtier);
}

struct cow_ops cow_mem_ops = {
	.cow_init     = cow_mem_init,
	.cow_push     = cow_mem_push,
	.cow_pull     = cow_mem_pull,
	.cow_log      = NULL,
	.cow_wait     = NULL,
	.cow_shutdown = NULL,
	.cow_deinit   = cow_mem_deinit,
	.cow_name     = "mem"
};
