/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include <linux/module.h>

#include "debug.h"
#include "export.h"
#include "filter.h"
#include "stcvsm_version.h"
#include "table.h"
#include "vdiff.h"
#include "portability.h"

int deny_export;

void *stcvsm_export_vdiff(unsigned int cmd)
{
	int i;
	int vdiff_idx;
	int num_vdiffs;
	struct exported_state *state;
	struct block_device_filter *table = stcvsm_filter_table();

	switch (cmd) {
	case 0:  /* start transaction */
		stcvsm_notice(LOG_MODULE, "start export transaction\n");
		if (deny_export)
			return ERR_PTR(-EACCES);
		if (quiesce_filters(1)) {
			quiesce_filters(0);
			return ERR_PTR(-EBUSY);
		}

		num_vdiffs = 0;
		for (i = 0; i < STCVSM_MAX_FILTERS; ++i)
			num_vdiffs += table[i].num_vdiffs;

		state = kmalloc(sizeof(struct exported_state) +
				sizeof(struct exported_filter) * num_vdiffs, GFP_KERNEL);
		if (state == NULL) {
			quiesce_filters(0);
			return ERR_PTR(-ENOMEM);
		}
		state->stcvsm_version = (STCVSM_MAJOR << 24) | (STCVSM_MINOR << 16) |
			(STCVSM_PATCH << 8) | (STCVSM_BUILD & 0xff);
		state->num_vdiffs = num_vdiffs;
		state->reserved1 = 0;
		state->reserved2 = 0;
		state->stcvsm_ioctl_vdiff = stcvsm_ioctl_vdiff;

		num_vdiffs = 0;
		for (i = 0; i < STCVSM_MAX_FILTERS; ++i) {
			for (vdiff_idx = 0; vdiff_idx < table[i].num_vdiffs; ++vdiff_idx) {
				struct vdiff_context *vdiff = table[i].vdiffs[vdiff_idx];

				state->filters[num_vdiffs].major = MAJOR(vdiff->vd_dev);
				state->filters[num_vdiffs].minor = MINOR(vdiff->vd_dev);
#ifdef HAVE_BDOPS_SUBMIT_BIO
				state->filters[num_vdiffs].orig_fops = table[i].bdf_orig_fops;
#else
				state->filters[num_vdiffs].orig_reqfn = table[i].bdf_orig_reqfn;
#endif
				num_vdiffs++;
			}
		}
		state->filters[num_vdiffs].major = 0;
		state->filters[num_vdiffs].minor = 0;
#ifdef HAVE_BDOPS_SUBMIT_BIO
		state->filters[num_vdiffs].orig_fops = NULL;
#else
		state->filters[num_vdiffs].orig_reqfn = NULL;
#endif

		return state;
	case 1:  /* commit transaction */
		stcvsm_notice(LOG_MODULE, "commit export transaction\n");
		for (i = 0; i < STCVSM_MAX_FILTERS; ++i)
		{
			spin_lock(bdf_lock(&table[i]));
			deactivate_filter(&table[i]);
			spin_unlock(bdf_lock(&table[i]));
		}
		shutdown_filters();
		return NULL;
	case 2:  /* cancel transaction */
		stcvsm_notice(LOG_MODULE, "cancel export transaction\n");
		quiesce_filters(0);
		return NULL;
	default:
		return ERR_PTR(-EINVAL);
	}
}
