/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include "stcvsm_os.h"
#include "look_aside_list.h"


/*
 * This lookaside module is provided as an interface for memory allocation.
 *
 * It is intended as an abstract interface for a possible a future implementation of slab or mempool
 * allocation in replacement of the current simple kalloc/kfree implementation of the code below.
 *
 * An earlier attempt at using slab allocation caused an OOPS in the kernelk while the
 * stcvsm-regression test ran. It seems there is a lazy resource re-claiming code path in removing
 * entries in the /proc/slabinfo name space.  That slabinfo name space has a requirement that there
 * be NO duplicate names in the name space.  stcvsm-regression loads, unloads and reloads the module
 * hundreds of time during it course of operation and it appears a duplication name is added to the
 * name space while a prior instance name still persists in the symbol table.
 *
 * In additon,  stcvsm_linux module has a mechanism to be replaced with a newer version without
 * losing state.  It is possible, due to the lazy resource freeing, that using the slab allocation
 * could also cause an OOPS by module re-loading. Be aware of this for future consideration.
 */

void lookaside_initialize(struct lookaside_list *lookaside,
				void *allocate, void *free,
				u32 flags,
				size_t size,
				u32 tag,
				u16 depth)
{
	/* stop unreferenced variable warning */
	(void) allocate;
	(void) free;
	(void) flags;
	(void) tag;
	(void) depth;
	lookaside->pool = NULL;
	lookaside->name = NULL;
	lookaside->cache = NULL;
	lookaside->size = size;
}

void *lookaside_allocate(struct lookaside_list *lookaside)
{
	return kmalloc(lookaside->size, GFP_KERNEL);
}

void lookaside_free(struct lookaside_list *lookaside, void *entry)
{
	(void) lookaside;  /* stop unreferenced variable warning */
	kfree(entry);
}

void lookaside_delete(struct lookaside_list *lookaside)
{
	(void) lookaside;  /* stop unreferenced variable warning */
}
