/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include "debug.h"
#include "portability.h"
#include "uassert.h"

struct mutex stcvsm_freeze_lock;

struct block_device *stcvsm_open_blkdev(dev_t dev, fmode_t mode, void *holder)
{
	struct block_device *bdev;

#ifdef HAVE_BLKDEV_GET_BY_DEV
	bdev = blkdev_get_by_dev(dev, mode, holder);
	if (IS_ERR(bdev)) {
		stcvsm_warning(LOG_MODULE, "Failed (%d,%d): %ld\n", MAJOR(dev), MINOR(dev),
			PTR_ERR(bdev));
	}
#else
	int err = 0;
	bdev = bdget(dev);
	if (bdev) {
		err = stcvsm_open_bdev(bdev, mode, holder);
		if (err == 0) {
			stcvsm_notice(LOG_REFS, "(%pg)(%d,%d)\n", bdev,
				atomic_read(&bdev->bd_inode->i_count), bdev->bd_openers);
		} else {
			bdev = NULL;
		}
	} else {
		err = -ENOENT;
	}
	if (!bdev) {
		stcvsm_warning(LOG_MODULE, "Failed (%d,%d): %d\n", MAJOR(dev), MINOR(dev), err);
		return ERR_PTR(err);
	}
#endif
	return bdev;
}

#ifndef HAVE_BLKDEV_GET_BY_DEV
int stcvsm_open_bdev(struct block_device *bdev, fmode_t mode, void *holder)
{
	int err;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 38)
	err = blkdev_get(bdev, mode, holder);
#else
	err = blkdev_get(bdev, mode & ~FMODE_EXCL);
	if (err == 0 && (mode & FMODE_EXCL) && holder) {
		if (bd_claim(bdev, holder) != 0) {
			err = -EBUSY;
			blkdev_put(bdev, mode);
		}
	}
#endif
	if (err < 0)
		return err;
	stcvsm_notice(LOG_REFS, "(%pg)(%d,%d)\n", bdev, atomic_read(&bdev->bd_inode->i_count),
		bdev->bd_openers);
	return 0;
}
#endif

void stcvsm_close_blkdev(struct block_device *bdev, fmode_t mode)
{
	stcvsm_notice(LOG_REFS, "(%pg)(%d,%d)\n", bdev, atomic_read(&bdev->bd_inode->i_count),
		bdev->bd_openers);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 38)
	if (mode & FMODE_EXCL) {
		bd_release(bdev);
		mode &= ~FMODE_EXCL;
	}
#endif
	blkdev_put(bdev, mode);
}

static inline struct super_block *__freeze_bdev(struct block_device *bdev)
{
	struct super_block *sb;
#ifdef HAVE_INT_FREEZE_BDEV
	int err = 0;
	// Is this a valid assumption anymore?
	err = freeze_bdev(bdev);
	if (err) {
		stcvsm_warning(LOG_SNAPSHOT, "Freeze failed for %pg (%d)", bdev, err);
	}
	sb = bdev->bd_fsfreeze_sb;
#else
	sb = freeze_bdev(bdev);
#endif
	return sb;
}

static inline int __thaw_bdev(struct block_device *bdev, struct super_block *sb)
{
	int err = 0;
#ifdef HAVE_INT_FREEZE_BDEV
	(void)sb;
	err = thaw_bdev(bdev);
#else
	err = thaw_bdev(bdev, sb);
#endif
	return err;
}

int stcvsm_freeze_bdev(struct block_device *bdev, struct super_block **sb)
{
	int err = 0;

	if (ERROR_INJECT(ERRINJ_BLKDEV_FREEZE))
		*sb = ERR_PTR(-EBUSY);
	else
		*sb = __freeze_bdev(bdev);
	if (!*sb || IS_ERR(*sb)) {
		if (!*sb) {
			/* Even if there was no superblock to freeze, freeze_bdev incremented the
			 * bdev->bd_fsfreeze_count.  Not clear if this is a bug or by design.
			 * Thaw (which does error check for NULL sb) to fix the refcount.
			 *
			 * Note that this has a dangerous race window due to the kernel bug:  If
			 * another thread calls freeze_bdev before I thaw it here, then it sees
			 * bd_fsfreeze_count > 0, and assumes the superblock exists, and crashes.
			 */
			__thaw_bdev(bdev, *sb);
		} else {
			err = PTR_ERR(*sb);
			stcvsm_warning(LOG_MODULE, "Failed (%pg): %d\n", bdev, -err);
			*sb = NULL;
		}
	} else {
		stcvsm_notice(LOG_MODULE, "Frozen (%pg)\n", bdev);
	}
	return err;
}

int stcvsm_thaw_bdev(struct block_device *bdev, struct super_block *sb)
{
	/* As long as we always call stcvsm_freeze_blkdev, a null super_block should have already
	 * had thaw called.  Calling it a second time would mean that bd_fsfreeze_count > 0, and
	 * assumes the super_block exists.  See stcvsm_freeze_blkdev
	 */
	int err;

	err = __thaw_bdev(bdev, sb);
	if (err < 0) {
		stcvsm_warning(LOG_MODULE, "Failed (%pg): %d\n", bdev, -err);
		return err;
	}
	stcvsm_notice(LOG_MODULE, "Thawed (%pg)\n", bdev);
	return 0;
}

static inline void stcvsm_partition_remap(struct bio *bio, struct block_device *bdev)
{
#if defined HAVE_BIO_REMAPPED
	if (!bio_flagged(bio, BIO_REMAPPED)) {
		bio_sector(bio) += get_start_sect(bdev);
		bio_set_flag(bio, BIO_REMAPPED);
	}
#elif defined HAVE_BIO_GENDISK
	if (bio->bi_partno) {
		bio_sector(bio) += get_start_sect(bdev);
		bio->bi_partno = 0;
	}
#else
	if (bdev != bdev->bd_contains) {
		bio_sector(bio) += get_start_sect(bdev);
		stc_bio_assign_disk(bio, bdev->bd_contains);
	}
#endif
}

MAKE_REQUEST_RETTYPE stcvsm_direct_submit_bio(struct bio *bio, struct block_device *bdev)
{
#if defined WITH_MQ_SUPPORT && defined MAKE_REQUEST_RETVAL
	MAKE_REQUEST_RETTYPE retval = MAKE_REQUEST_RETVAL;
#endif
	struct gendisk *disk = bdev->bd_disk;
	struct request_queue *q = disk->queue;
	/* count_vm_events: Skipping.  "sar pgpgout" will be inaccurate.
	 * Yet md and drbd skip this by going directly to
	 * generic_make_request, so already PGPGOUT only describes the IO
	 * initiated higher in the stack, not the actual IO that hits the
	 * driver.
	 */

	/* generic_make_request works hard to not call submit_bio_fn recursively.
	 * We know we're already called once on this task, but potentially
	 * on a different queue.
	 *
	 */

	/* blk_partition_remap */
	stcvsm_partition_remap(bio, bdev);

	/* bio_check_eod: cow_pack_bios is trusted to not try to write past
	 * the end of the disk.
	 *
	 * TODO Test COW to LVM, and lvreduce it.
	 */

	/* CONFIG_FAIL_MAKE_REQUEST: Not supporting fault injection.  It's
	 * your lucky day.
	 */

	/* Bio is not a FLUSH, DISCARD, or WRITE_SAME; no special handling */

	/* create_io_context: Only needs to be called once per task, so
	 * we're good.  This can fail anyway, and supposedly the block
	 * layer knows how to cope.
	 */

	/* blkcg_bio_issue_check: This throttles bios in a cgroup.  The
	 * original write was already throttled.  COW writes are a
	 * by-product.  Skip.
	 */

#ifdef WITH_MQ_SUPPORT
	// See do_make_request/__do_submit_bio
	percpu_ref_get(&q->q_usage_counter);
	if (!DISK_SUBMIT_BIO(disk))
		return stcvsm_mq_submit_bio(bio);
#ifdef MAKE_REQUEST_RETVAL
	retval = stcvsm_release_bio(bio, DISK_SUBMIT_BIO(disk));
#else
	stcvsm_release_bio(bio, DISK_SUBMIT_BIO(disk));
#endif
	percpu_ref_put(&q->q_usage_counter);
	return retval;
#else
	(void)q;
	return stcvsm_release_bio(bio, DISK_SUBMIT_BIO(disk));
#endif
}

#ifdef WITH_MQ_SUPPORT
#ifdef HAVE_BLK_MQ_MAKE_REQUEST
blk_qc_t stcvsm_sq_to_mq(struct request_queue *q, struct bio *bio)
{
	percpu_ref_get(&q->q_usage_counter);
	return stcvsm_mq_submit_bio(bio);
}
#else
blk_qc_t stcvsm_sq_to_mq(struct bio *bio)
{
	struct request_queue *q = bio_request_queue(bio);
	percpu_ref_get(&q->q_usage_counter);
	return stcvsm_mq_submit_bio(bio);
}
#endif
#endif

spinlock_t* __q_lock(struct request_queue *q)
{
#ifdef HAVE_RQ_Q_NO_PTR
	return &q->queue_lock;
#else
	return q->queue_lock;
#endif
}

void stcvsm_lock_queue(struct request_queue *q)
{
	spin_lock_irq(__q_lock(q));
}

void stcvsm_unlock_queue(struct request_queue *q)
{
	spin_unlock_irq(__q_lock(q));
}

#ifndef HAVE_BLK_ALLOC_DISK
#ifdef HAVE_BDOPS_SUBMIT_BIO
struct request_queue *stcvsm_snap_alloc_queue(struct block_device *bdev)
{
	return blk_alloc_queue(NUMA_NO_NODE);
}
#else
struct request_queue *stcvsm_snap_alloc_queue(struct block_device *bdev, submit_bio_fn *request_fn)
{
	struct request_queue *q = NULL;
#ifdef HAVE_BLK_ALLOC_QUEUE_RQFN
	q = blk_alloc_queue(request_fn, NUMA_NO_NODE);
#elif defined HAVE_BLK_ALLOC_GFP_T
	q = blk_alloc_queue(GFP_KERNEL);
#else
#error How to allocate a snapshot queue
#endif
	if(q)
	{
#ifdef HAVE_BLK_ALLOC_GFP_T
		blk_queue_make_request(q, request_fn);
#endif
#if defined HAVE_BLK_ALLOC_GFP_T || defined HAVE_BLK_ALLOC_QUEUE_RQFN
		blk_queue_stack_limits(q, bdev->bd_disk->queue);
#endif
	}
	return q;
}
#endif
#endif // HAVE_BLK_ALLOC_DISK

MAKE_REQUEST_RETTYPE stcvsm_release_bio(struct bio *bio, submit_bio_fn *fn)
{
#ifdef HAVE_BDOPS_SUBMIT_BIO
	return fn(bio);
#else
	return fn(bio_request_queue(bio), bio);
#endif
}

#ifdef WITH_MQ_SUPPORT
MAKE_REQUEST_RETTYPE stcvsm_mq_submit_bio(struct bio *bio)
{
#ifdef HAVE_BLK_MQ_MAKE_REQUEST
	return blk_mq_make_request(bio_request_queue(bio), bio);
#else
#ifndef HAVE_BLK_MQ_SUBMIT_BIO
#error "Missing address for blk_mq_submit_bio"
#endif
	static const submit_bio_fn *__blk_mq_submit_bio =
		(const submit_bio_fn*)(BLK_MQ_SUBMIT_BIO_ADDR + kfree - KFREE_ADDR);
	return __blk_mq_submit_bio(bio);
#endif
}
#endif
