/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/string.h>
#include <linux/uaccess.h>

#include "cow.h"
#include "debug.h"
#include "filter.h"
#include "module.h"
#include "proc.h"
#include "snapshot.h"
#include "stcvsmctl.h"
#include "stcvsm_version.h"
#include "table.h"
#include "vdiff.h"

#include "stcsnap_p.h"
#include "stcsnapdef_p.h"

extern int deny_export;
static struct proc_dir_entry *stcvsm_dir;

static int stcvsm_config_read(struct seq_file *s, void *data)
{
	int i;
	struct block_device_filter *table = stcvsm_filter_table();

	seq_printf(s, "version: %s\n", STCVSM_VERSION);
	seq_printf(s, "stcvsm_major: %d\n", stcvsm_major);
	seq_printf(s, "deny_export: %d\n", deny_export);

	seq_puts(s, "filters:");
	for (i = 0; i < STCVSM_MAX_FILTERS; ++i) {
		struct block_device_filter *bdfilter = &table[i];

		if (atomic_read(&bdfilter->bdf_refs)) {
			/*
			 * Must grab the lock and re-verify to avoid using partially constructed
			 * or deconstructed filter.
			 */
			dev_t dev = 0;

			spin_lock(bdf_lock(bdfilter));
			if (atomic_read(&bdfilter->bdf_refs))
				dev = bdfilter->bdf_bdev->bd_dev;
			spin_unlock(bdf_lock(bdfilter));
			if (dev)
				seq_printf(s, " %u:%u", MAJOR(dev), MINOR(dev));
		}
	}
	seq_puts(s, "\n");

	return 0;
}

static ssize_t stcvsm_config_write(struct file *file, const char __user *buf, size_t count,
		loff_t *ppos)
{
	int err;
	unsigned int major, minor;
	char buffer[64];

	if (!buf || count >= sizeof(buffer))
		return -EINVAL;

	/* The min() is to pacify gcc */
	if (copy_from_user(buffer, buf, min(sizeof(buffer), count)))
		return -EFAULT;
	buffer[count] = '\0';

	if (sscanf(buffer, "vdiff stop %u:%u", &major, &minor) == 2) {
		stcvsm_debug(LOG_PROC, "vdiff stop %u:%u\n", major, minor);
		err = stcvsm_ioctl_vdiff(MKDEV(major, minor), IOCTL_VSNAP_VDIFF_STOP, NULL);
	} else if (sscanf(buffer, "snapshot stop %u:%u", &major, &minor) == 2) {
		stcvsm_debug(LOG_PROC, "snapshot stop %u:%u\n", major, minor);
		err = stcvsm_ioctl_snapshot(MKDEV(major, minor), IOCTL_UNSNAP_VOLUME, NULL);
	} else {
		return -EINVAL;
	}

	/* Specifically treating ENOENT as success for "stop" actions.  (It is gone, as desired.)
	 * Common use case is a shell script to stop all snapshots, but stopping one in a set stops
	 * the entire set.
	 */
	if (err == 0 || err == -ENOENT)
		err = count;

	return err;
}

static int filter_proc_read(struct seq_file *s, void *data)
{
	int idx;
	int refs, num_snapshots, num_vdiffs;
	struct snapshot_context **snapshots;
	struct snapshot_context *snapshot;
	struct vdiff_context *vdiff;
	struct block_device_filter *bdfilter = (struct block_device_filter *) s->private;
	dev_t dev;

	/* Take care to not hold the lock over seq_* calls, which can kmalloc.  Instead pin all
	 * snapshots and vdiffs at once before starting.
	 *
	 * vdiff will stay in the array while the refcount is nonzero.  But snapshots can be pulled
	 * out of the array early, to avoid new references.  Therefore must copy the snapshot array.
	 */
	spin_lock(bdf_lock(bdfilter));
	num_snapshots = bdfilter->num_snapshots;
	spin_unlock(bdf_lock(bdfilter));

	snapshots = kmalloc_array(num_snapshots, sizeof(struct snapshot_context *), GFP_KERNEL);
	if (snapshots == NULL)
		return -ENOMEM;

	spin_lock(bdf_lock(bdfilter));
	refs = atomic_read(&bdfilter->bdf_refs);
	if (refs == 0) {
		dev = 0;
		num_snapshots = 0;
		num_vdiffs = 0;
	} else {
		dev = bdfilter->bdf_bdev->bd_dev;
		if (num_snapshots > bdfilter->num_snapshots)
			num_snapshots = bdfilter->num_snapshots;
		for (idx = 0; idx < num_snapshots; ++idx) {
			snapshots[idx] = bdfilter->snapshots[idx];
			stcvsm_atomic_inc(&snapshots[idx]->sc_refs);
		}
		num_vdiffs = bdfilter->num_vdiffs;
		for (idx = 0; idx < num_vdiffs; ++idx) {
			vdiff = bdfilter->vdiffs[idx];
			stcvsm_atomic_inc(&vdiff->vd_refs);
		}
	}
	spin_unlock(bdf_lock(bdfilter));

	seq_printf(s,
			"filter %d\n"
			" refs %d\n",
			bdfilter->idx,
			refs);
	if (refs == 0) {
		kfree(snapshots);
		return 0;
	}
	seq_printf(s, " filter dev_t %u:%u\n", MAJOR(dev), MINOR(dev));

	for (idx = 0; idx < num_snapshots; ++idx) {
		snapshot = snapshots[idx];
		dev = snapshot->sc_bdev->bd_dev;
		seq_printf(s,
				"snapshot %d\n"
				" refs %d\n"
				" state %s\n"
				" set %pUL\n"  /* STC GUIDs are always LE capitalized */
				" original dev_t %u:%u\n"
				" snapshot dev_t %u:%u\n",
				idx,
				atomic_read(&snapshot->sc_refs),
				snapshot_state_to_str(snapshot->sc_flags),
				snapshot->set_guid,
				MAJOR(dev), MINOR(dev),
				stcvsm_major, snapshot->sc_snap_minor);
		cow_log_stats(snapshot->cstor, s);
	}
	for (idx = 0; idx < num_vdiffs; ++idx) {
		vdiff = bdfilter->vdiffs[idx];
		dev = vdiff->vd_dev;
		seq_printf(s,
				"vdiff %d\n"
				" refs %d\n"
				" chain %pUL\n"
				" file %pUL\n"
				" vdiff dev_t %u:%u\n"
				" dirty sectors %llu of %llu\n",
				idx,
				atomic_read(&vdiff->vd_refs),
				vdiff->VDiffGuid,
				vdiff->VDiffFileGuid,
				MAJOR(dev), MINOR(dev),
				(unsigned long long)count_vdiff(vdiff),
				(unsigned long long)vdiff->vd_sectors
				);
	}

	/* Free vdiffs backwards so the array is guaranteed to stay stable */
	for (idx = num_vdiffs - 1; idx >= 0; --idx)
		put_vdiff(bdfilter->vdiffs[idx]);
	for (idx = 0; idx < num_snapshots; ++idx)
		put_snapshot(snapshots[idx]);
	kfree(snapshots);

	return 0;
}

static int __init stcvsm_init_proc_dir(void)
{
	stcvsm_dir = proc_mkdir("stcvsm", NULL);
	if (!stcvsm_dir)
		return -ENOMEM;
	return 0;
}

static void stcvsm_remove_proc_dir(void)
{
	remove_proc_entry("stcvsm", NULL);
	stcvsm_dir = NULL;
}

static int stcvsm_config_open(struct inode *inode, struct file *file)
{
	return single_open(file, stcvsm_config_read, NULL);
}

#ifdef HAVE_PROC_OPS
static const struct proc_ops config_proc_fops = {
	.proc_open = stcvsm_config_open,
	.proc_read = seq_read,
	.proc_lseek = seq_lseek,
	.proc_write = stcvsm_config_write,
	.proc_release = single_release,
};
#else
// Kernel 5.8 changed the interface to proc_create_data to take a `struct proc_ops`
static const struct file_operations config_proc_fops = {
	.owner = THIS_MODULE,
	.open = stcvsm_config_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.write = stcvsm_config_write,
	.release = single_release,
};
#endif

static int __init stcvsm_init_proc_config(void)
{
	if (!proc_create_data("config", 0644, stcvsm_dir, &config_proc_fops, NULL))
		return -ENOMEM;

	return 0;
}

static void stcvsm_remove_proc_config(void)
{
	remove_proc_entry("config", stcvsm_dir);
}

static int filter_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, filter_proc_read, PDE_DATA(inode));
}

#ifdef HAVE_PROC_OPS
static const struct proc_ops filter_proc_fops = {
	.proc_open = filter_proc_open,
	.proc_read = seq_read,
	.proc_lseek = seq_lseek,
	.proc_release = single_release,
};
#else
static const struct file_operations filter_proc_fops = {
	.owner = THIS_MODULE,
	.open = filter_proc_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};
#endif

static char *filter_proc_name(char *str, dev_t dev)
{
	scnprintf(str, 16, "%u:%u", MAJOR(dev), MINOR(dev));
	return str;
}

int stcvsm_init_proc_filter(struct block_device_filter *bdfilter)
{
	struct proc_dir_entry *entry;
	char str[16];

	entry = proc_create_data(filter_proc_name(str, bdfilter->bdf_bdev->bd_dev), 0644,
			stcvsm_dir, &filter_proc_fops, bdfilter);
	if (!entry)
		return -ENOMEM;

	return 0;
}

void stcvsm_remove_proc_filter(dev_t dev)
{
	char str[16];

	remove_proc_entry(filter_proc_name(str, dev), stcvsm_dir);
}

int __init stcvsm_init_proc(void)
{
	int err;

	err = stcvsm_init_proc_dir();
	if (err < 0)
		return err;

	err = stcvsm_init_proc_config();
	if (err < 0) {
		stcvsm_remove_proc_dir();
		return err;
	}

	return 0;
}

void stcvsm_remove_proc(void)
{
	stcvsm_remove_proc_config();
	stcvsm_remove_proc_dir();
}
