/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

/*
 * Utility package - page-based range table
 * 64bit ready
 */
#ifndef STCVSM_RANGE_TABLE_H
#define STCVSM_RANGE_TABLE_H

#include "tree.h"


/* Constants */

/*
 * Page size
 * Includes some space for tree item header
 */
#define RANGE_PAGE_SIZE                      4000

/* Maximum number of entries per page */
#define PAGE_RANGE_MAX      ((RANGE_PAGE_SIZE - sizeof(struct page_of_ranges)) / sizeof(URPVALUE))

/*
 * Merge threshold - 2 pages are merged if total number of items in them is <= this value
 * 2/3 of a full page
 */
#define PAGE_RANGE_MERGE_THRESHOLD           ((PAGE_RANGE_MAX * 2) / 3)


/*
 * Range table contains a set of [N..M) ranges
 * They are automatically coalesced to save memory, and their boundaries are not
 *  preserved
 * So, the table only keeps the fact what positions are covered and what are not
 * This table implementation (unlike the one in urange.h/.c) keeps the ranges as linear
 *  arrays within the pages, and the pages themselves - in the AVL tree
 */


/* Types */

/* Type of range table value */
typedef u32 URPVALUE, *PURPVALUE;

/* Structures */

/*
 * Flag values for page_of_ranges_find_page
 * Treat the page's first value as Inside (otherwise - as Below/Between)
 */
#define RANGE_SEARCH_WITH_START                 1
/* Treat the page's last value as Inside (otherwise - as Between/Above) */
#define RANGE_SEARCH_WITH_END                   2

/* Search results from page_of_ranges_find_page */
enum enum_search_result {
	/* Below the first page */
	RangeTreeBelow,
	/* Inside some page */
	RangeTreeInside,
	/* Between the pages */
	RangeTreeBetween,
	/* Above the last page or the tree is empty */
	RangeTreeAbove
};


/*
 * Range page itself
 * Always allocated as RANGE_PAGE_SIZE bytes, this structure is the header
 * Allocation must always provide the alignment of sizeof(URPVALUE)
 */
struct page_of_ranges {
	/*
	 * Base value for ranges
	 * Added to all "wall" values to obtain the real range start/end
	 */
	u64 Base;
	/*
	 *  Number of valid "walls" follow
	 * Must be at least 2
	 */
	u32 Count;
	/*
	 * URANPAGE_USE_CRC
	 * Number of values in the page
	 */
	URPVALUE ValuesCount;
	/*
	 * URPVALUE-aligned here
	 * Sorted array of URPVALUE "walls" follow
	 */
};

/* Page template for tree insert and lookup */
struct page_of_ranges_template {
	/* Header */
	struct page_of_ranges Page;
	/* First value */
	URPVALUE Value0;
};

/* A range and a reference to its location within the table */
struct range_ref {
	/* Start of the range */
	u64 Start;
	/* End of the range
	 * Same as Start if this structure is logical NULL
	 */
	u64 End;
	/* Page
	 * NULL if this structure is logical NULL
	 */
	void *Page;
	/* Index within page */
	u32 Index;
};

/* Page-based range table */
struct range_page_table {
	/* Tree (of pages) */
	struct tree_table Tree;
	/* Number of ranges in the table */
	u32 RangesCount;
	/* Number of values covered by the table ranges - the summary length of all table ranges */
	u64 ValuesCount;
};

/*
 * Context for range_enumerate_merge_adjacent
 * Opaque, functions and macros must be used to access
 */
struct range_page_table_enum_context {
	/* Next table range */
	struct range_ref Range;
	/* Start of accumulated range */
	u64 Start;
	/* End of accumulated range */
	u64 End;
};

/* Return value of range_enumber_merge_adjacent */
struct range_page_table_enum_retval {
	/* Start of range */
	u64 Start;
	/* End of range */
	u64 End;
};

/* Macros */

/* Range count in the table */
#define RangeCount(_T)			((_T)->RangesCount)

/* Unique values covered by ranges count in the table */
#define RangeValuesCount(_T)		((_T)->ValuesCount)

/* Initializes struct range_page_table_enum_context */
#define RangeEnumerateContextInit(_C)	(_C)->Range.Page = (void *)(s64)(-1)

/* Sets the End value and Count in a page of 1 range */
#define PAGE_SET_END(_P, _E)                    do { \
		(_P)->Count = 2; \
		((PURPVALUE)((_P)+1))[1] = (_E); \
} while (0)
/* Page's start value
 * L-value
 */
#define PAGE_START(_P)                          ((PURPVALUE)((_P)+1))[0]
/* Page's start value (absolute) */
#define PAGE_START_ABS(_P)                      ((_P)->Base + PAGE_START(_P))
/* Page's second-to-start value
 * L-value
 */
#define PAGE_START2(_P)                         ((PURPVALUE)((_P)+1))[1]
/* Page's end value
 * L-value
 */
#define PAGE_END(_P)                            ((PURPVALUE)((_P)+1))[(_P)->Count - 1]
/* Page's end value (absolute) */
#define PAGE_END_ABS(_P)                        ((_P)->Base + PAGE_END(_P))
/* End of valid page data */
#define PAGE_DATA_END(_P)                       (&(((PURPVALUE)((_P)+1))[(_P)->Count]))


/* Functions */

/* Initializes a nonpaged table */
void range_initialize_nonpaged(struct range_page_table *table, enum tree_kind TreeKind,
			       struct lookaside_list *lookaside, u32 PoolTag);

/*
 * Cleans up a table, deleting all items and cleaning up the lookaside allocator
 * Both allocator and a table can be reused then
 */
void range_cleanup(struct range_page_table *table);
/*
 * Inserts a range to the range page table
 * Returns false on allocation failure, the table is not updated in this case
 * The caller must guarantee Start < End
 */
bool range_insert(struct range_page_table *table,
		     u64 Start, u64 End);
/*
 * Deletes a range from the range page table
 * Returns false on allocation failure (if hole carving was tried and failed),
 *  the table is not updated in this case
 * The caller must guarantee Start < End
 */
bool range_delete(struct range_page_table *table,
		     u64 Start, u64 End);
/*
 * Deletes a range which is in the table by its reference (was found by range_hit or
 *  range_enumerate, or obtained by range_next/Prev/First/Last)
 */
void range_delete_existing(struct range_page_table *table, struct range_ref *Range);
/* Deletes all ranges from the table */
void range_delete_all(struct range_page_table *table);
/*
 * Looks up whether Value is covered by some range
 * If not covered - returns false and sets *Range to reference the next table range just
 *  after Value, or to logical NULL if the table is empty or if Value is > then
 *  any table range
 * If covered - returns true and sets *Range to reference the range which contains
 *  this value
 * The returned range reference is valid till next table update or cleanup
 */
bool range_hit(struct range_page_table *table, u64 Value,
		  struct range_ref *Range);
/*
 * Advances *Range to the next range or to logical NULL if the given range is last
 * Can return adjacent ranges
 */
void range_next(struct range_ref *Range);
/*
 * Advances *Range to the previous range or to logical NULL if the given range is first
 * Can return adjacent ranges
 */
void range_prev(struct range_ref *Range);
/*
 * Sets *Range to reference the first range in the table
 *  or to logical NULL if the table is empty
 * Can return adjacent ranges
 */
void range_first(struct range_page_table *table, struct range_ref *Range);
/*
 * Sets *Range to reference the last range in the table
 *  or to logical NULL if the table is empty
 * Can return adjacent ranges
 */
void range_last(struct range_page_table *table, struct range_ref *Range);
/*
 * Table enumerator
 * Returns ranges in the comparison order
 * Does not do any reorders to the table
 * Returns (via *Range) the next range if found or logical NULL if no more items
 * Returns true if the next range was found, false otherwise
 * Range->Page should be set to (void *)(64)(-1) before calling this for the first time
 *  to initialize the enumeration context (Range)
 * Can return adjacent ranges
 */
bool range_enumerate(struct range_page_table *table, struct range_ref *Range);
/*
 * Same as range_enumerate but merges the adjacent ranges
 * Context must be initialized by range_enumerate_context_init before the first call,
 *  and reused for next calls
 * Return value is in RetVal
 */
bool range_enumerate_merge_adjacent(struct range_page_table *table,
				     struct range_page_table_enum_context *Context,
				     struct range_page_table_enum_retval *RetVal);
/*
 * Looks up whether the range is completely contained in the table
 * Returns true if contained
 */
bool range_check_for_dense_range(struct range_page_table *table,
				 u64 Start, u64 End);
/*
 * Looks up whether the range is completely not contained in the table
 * Returns true if not contained
 */
bool range_check_for_hole_range(struct range_page_table *table,
				u64 Start, u64 End);
/*
 * Checks whether [Start..End) intersects with any in-table range
 * Returns true if yes, and fills the intersection (least of them if there are several)
 *  to *Range, the table location fields in *Range are not defined
 * Returns false if no, the state of *Range is not defined in this case
 */
bool range_next_range_in(struct range_page_table *table,
			  u64 Start, u64 End,
			  struct range_ref *Range);
/*
 * Checks whether [Start..End) intersects with any hole between in-table ranges
 * Returns true if yes, and fills the intersection (least of them if there are several)
 *  to *Range, the table location fields in *Range are not defined
 * Returns false if no, the state of *Range is not defined in this case
 */
bool range_next_range_out(struct range_page_table *table,
			   u64 Start, u64 End,
			   struct range_ref *Range);

#ifdef UNITTEST_CODE
extern bool range_check(struct range_page_table *table);
#endif  /* UNITTEST_CODE */

#endif /* STCVSM_RANGE_TABLE_H */
