/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include <linux/genhd.h>
#include <linux/bio.h>
#include <linux/hdreg.h>
#include <linux/module.h>

#include "cow.h"
#include "debug.h"
#include "uassert.h"
#include "ext2_bitmap.h"
#include "filter.h"
#include "module.h"
#include "portability.h"
#include "snapshot.h"
#include "workqueue.h"
#include "vdiff.h"

#ifdef WITH_EXT2BITMAP
static int stcvsm_fs_bitmap(struct snapshot_context *snap, struct super_block *sb);
#endif
#ifdef HAVE_BDOPS_SUBMIT_BIO
static MAKE_REQUEST_RETTYPE snap_submit_bio(struct bio *bio);
#else
static MAKE_REQUEST_RETTYPE stcvsm_snap_reqfn(struct request_queue *q, struct bio *bio);
#endif

struct cow_fill_job {
	struct cow_storage *cstor;
	struct cow_buffer *cow;
	int error;
	atomic_t num_bios;
};

struct cow_overlay_job {
	struct wq_work w;
	struct cow_storage *cstor;
	struct cow_buffer **cows;
	u8 *readers;
	unsigned int num_cows;
	struct bio *orig_bio;
#ifdef HAVE_HD_STRUCT_IN_FLIGHT
	struct hd_struct *hd;
#endif
	unsigned int start_time;
};

struct cow_writer_job {
	struct wq_work w;
	struct cow_buffer **cows;
	unsigned int num_cows;
	struct bio *orig_bio;
	struct snapshot_context *snap;
};

/* open */

static int snap_open(struct block_device *bdev, fmode_t mode)
{
	struct gendisk *disk = bdev->bd_disk;
	struct request_queue *q = disk->queue;
	struct snapshot_context *snap;
	struct block_device_filter *bdfilter = disk->private_data;
	int err = 0;

	spin_lock(bdf_lock(bdfilter));
	snap = q->queuedata;
	if (snap) {
		if (!test_bit(SNAPSHOT_INIT, &snap->sc_flags))
			err = -EAGAIN;
		else if (test_bit(SNAPSHOT_SHUTDOWN, &snap->sc_flags))
			err = -EINVAL;
		else
			stcvsm_atomic_inc(&snap->sc_refs);
	} else {
		err = -ENOENT;
		stcvsm_warning(LOG_SNAPSHOT, "(%pS) No queuedata: %d\n", snap, err);
	}
	spin_unlock(bdf_lock(bdfilter));

	if (!err)
		stcvsm_info(LOG_SNAPSHOT, "%d(%p)\n", atomic_read(&snap->sc_refs), snap);
	else
		stcvsm_warning(LOG_SNAPSHOT, "(%pS) error opening snapshot bd: %d\n", snap, err);
	return err;
}

/* release */

static void snap_release(struct gendisk *disk, fmode_t mode)
{
	struct snapshot_context *snap = disk->queue->queuedata;
	stcvsm_info(LOG_SNAPSHOT, "%d(%p)\n", atomic_read(&snap->sc_refs), snap);
	put_snapshot(snap);
}

/* ioctl */

static int snap_getgeo(struct block_device *bdev, struct hd_geometry *geometry)
{
	struct snapshot_context *snap = bdev->bd_disk->queue->queuedata;

	memset(geometry, 0, sizeof(*geometry));
	geometry->cylinders = (snap->block_size * snap->sectors_per_block) >> 6;
	geometry->heads = 4;
	geometry->sectors = 16;
	geometry->start = 0;
	return 0;
}

static int snap_ioctl(struct block_device *bdev, fmode_t mode, unsigned int cmd, unsigned long param)
{
	struct hd_geometry geo;

	stcvsm_info(LOG_SNAPSHOT, "snap_ioctl(%p,%d,%u,%lu)\n", bdev, mode, cmd, param);
	switch (cmd) {
	case HDIO_GETGEO:
		snap_getgeo(bdev, &geo);
		return copy_to_user((void __user *)param, &geo, sizeof(geo));
	}

	return -ENOTTY;
}

static struct block_device_operations snap_ops = {
	.owner			= THIS_MODULE,
#ifdef HAVE_BDOPS_SUBMIT_BIO
	.submit_bio		= snap_submit_bio,
#endif
	.open			= snap_open,
	.release		= snap_release,
	.ioctl			= snap_ioctl,
	.getgeo			= snap_getgeo
};

struct snapshot_context *get_snapshot_bio(struct block_device_filter *bdfilter, struct bio *bio)
{
	int idx;
	sector_t sector_start = bio_sector(bio);
	sector_t sector_end = sector_start + bio_sectors(bio);

	for (idx = 0; idx < bdfilter->num_snapshots; ++idx) {
		struct snapshot_context *snap = bdfilter->snapshots[idx];

		if (sector_start >= snap->sc_sector_start && sector_end <= snap->sc_sector_end &&
				test_bit(SNAPSHOT_SNAPPED, &snap->sc_flags)) {
			stcvsm_atomic_inc(&snap->sc_refs);
			return snap;
		}
	}
	return NULL;
}

struct snapshot_context *get_snapshot_bdev(struct block_device_filter *bdfilter, struct block_device *bdev)
{
	int idx;

	for (idx = 0; idx < bdfilter->num_snapshots; ++idx) {
		struct snapshot_context *snap = bdfilter->snapshots[idx];

		if (test_bit(SNAPSHOT_INIT, &snap->sc_flags) && snap->sc_bdev->bd_dev == bdev->bd_dev) {
			stcvsm_atomic_inc(&(snap->sc_refs));
			return snap;
		}
	}
	return NULL;
}

struct block_device *get_snapshot_filter_bdev(struct block_device *bdev)
{
	struct snapshot_context *snap = bdev->bd_disk->queue->queuedata;
	return snap->sc_bdev;
}

static int get_snapshot_minor(struct snapshot_context *snap)
{
	int i, err = -EMFILE;

	if (ERROR_INJECT(ERRINJ_SNAPSHOT_MINOR))
		return err;

	spin_lock(bdf_lock(snap->bdfilter));
	for (i = 0; i < MAX_SNAPS_PER_FILTER; ++i) {
		if (!(snap->bdfilter->snapshot_mask & (1U<<i))) {
			snap->bdfilter->snapshot_mask |= (1U<<i);
			snap->sc_snap_minor = snap->bdfilter->idx * 16 + i+1;
			err = 0;
			break;
		}
	}
	spin_unlock(bdf_lock(snap->bdfilter));
	return err;
}

static void put_snapshot_minor(struct block_device_filter *bdfilter, int minor)
{
	spin_lock(bdf_lock(bdfilter));
	bdfilter->snapshot_mask &= ~(1U<<((minor & 0xf)-1));
	spin_unlock(bdf_lock(bdfilter));
}

static void get_snapshot_name(char *name, int minor)
{
	scnprintf(name, DISK_NAME_LEN, "stcvsm_snap%d", minor);
	stcvsm_info(LOG_SNAPSHOT, "%s\n", name);
}

static void free_snapshot(struct snapshot_context *snap)
{
	cstor_destruct(snap->cstor);
	stcvsm_close_blkdev(snap->sc_bdev, FMODE_READ);
	kfree(snap->freespc_bitmap);
	kfree(snap);
	module_put(THIS_MODULE);
}

static struct snapshot_context *alloc_snapshot(struct block_device *bdev, dev_t cow_dev)
{
	int err;
	struct snapshot_context *snap;
	sector_t sectors;
	/* Don't want a block size smaller than the filesystem, due to overhead.
	 * Could be larger.  Mounted filesystems can't be larger than PAGE_SIZE.
	 */
	unsigned int block_size = PAGE_SIZE;

	if (!try_module_get(THIS_MODULE)) {
		stcvsm_notice(LOG_MODULE, "module is unloading");
		err = -EAGAIN;
		goto error0;
	}

	if (ERROR_INJECT(ERRINJ_ALLOC))
		snap = NULL;
	else
		snap = kzalloc(sizeof(*snap), GFP_KERNEL);
	if (!snap) {
		err = -ENOMEM;
		goto error1;
	}
	snap->sc_bdev = stcvsm_open_blkdev(bdev->bd_dev, FMODE_READ, NULL);
	if (IS_ERR(snap->sc_bdev)) {
		err = PTR_ERR(snap->sc_bdev);
		goto error2;
	}
	atomic_set(&snap->sc_refs, 1);
	snap->block_size = block_size;
	snap->sectors_per_block = block_size >> 9;
	sectors = i_size_read(snap->sc_bdev->bd_inode) >> 9;
	snap->bits_in_map = sectors + snap->sectors_per_block - 1;
	do_div(snap->bits_in_map, snap->sectors_per_block);
	stcvsm_debug(LOG_SNAPSHOT, "bits_in_map = %lu, block_size = %u\n", snap->bits_in_map,
			snap->block_size);
#ifdef WITH_EXT2BITMAP
	/* TODO: stcvsm_fs_bitmap(snap); */
	/* TODO: Move to activate_snapshots */
	(void)&stcvsm_fs_bitmap;
#endif
	snap->cstor = cstor_construct(sectors, bdev->bd_dev);
	if (!snap->cstor) {
		err = -ENOMEM;
		goto error3;
	}
	err = cstor_add_tier(snap->cstor, &cow_mem_ops);
	if (err)
		goto error4;
	/* Subsequent tiers are optional; no error check needed */
	if (cow_dev) {
		struct block_device *cow_bdev;

		cow_bdev = stcvsm_open_blkdev(cow_dev, FMODE_READ, NULL);
		if (!IS_ERR(cow_bdev)) {
			int sameq;

			/* Indicate that the snapshot and cow devices share the same request_queue
			 * to control throttling and direct queue submission. */
			sameq = (bdev_get_queue(snap->sc_bdev) == bdev_get_queue(cow_bdev));
			stcvsm_close_blkdev(cow_bdev, FMODE_READ);

			if (sameq)
				stcvsm_notice(LOG_SNAPSHOT, "Snapshot and cow device share a queue.");
			cstor_add_tier(snap->cstor, &cow_bdev_ops, COW_BDEV_DEDICATED,
				cow_dev, sameq);
		}
	} else if (snap->freespc_bitmap) {
		cstor_add_tier(snap->cstor, &cow_bdev_ops, COW_BDEV_FREESPACE, bdev->bd_dev,
				snap->bits_in_map, snap->freespc_bitmap);
	}
	snap->sc_sector_start = get_start_sect(snap->sc_bdev);
	snap->sc_sector_end   = snap->sc_sector_start + stcvsm_nr_sectors(snap->sc_bdev);
	snap->sc_snap_minor = -1; /* unallocated */
	return snap;
error4:
	cstor_destruct(snap->cstor);
error3:
	stcvsm_close_blkdev(snap->sc_bdev, FMODE_READ);
error2:
	kfree(snap);
error1:
	module_put(THIS_MODULE);
error0:
	return ERR_PTR(err);
}

/** Flush and destroy the snapshot block device.
 * In kernels 5.13+, bd_mutex is held while calling block_device_operations->release, and is
 * acquired during del_gendisk.
 * We'd rather keep the feature of destroying snapshot when the last file descriptor closes than
 * return EAGAIN or EBUSY from ioctls to stop the snapshot.
 */

struct shutdown_snapshot_job {
	struct wq_work w;
	struct gendisk *disk;
	struct request_queue *q;
	unsigned long flags;
	int minor;
	struct block_device_filter *bdfilter;
};

static void shutdown_snapshot_work(struct wq_work* work)
{
	struct shutdown_snapshot_job *job = (struct shutdown_snapshot_job*)work;

	if (test_bit(SNAPSHOT_INIT, &job->flags)) {
		del_gendisk(job->disk);
	} else {
		/* If snapshot was not fully initialized, add_disk() was never called.  Cleanup
		 * disk->queue before calling put_disk() so that we never try to put a reference
		 * which was never taken in first place.
		 */
		job->disk->queue = NULL;
	}

	blk_cleanup_queue(job->q);

	put_disk(job->disk);

	if (job->minor >= 0)
		put_snapshot_minor(job->bdfilter, job->minor);

	kfree(job);
}

static void shutdown_snapshot_blkdev(struct snapshot_context *snap)
{
	struct shutdown_snapshot_job *job = kzalloc(sizeof(struct shutdown_snapshot_job), GFP_NOFS);

	job->w.fn = shutdown_snapshot_work;
	job->q = snap->sc_snap_queue;
	job->disk = snap->sc_snap_gendisk;
	job->flags = snap->sc_flags;
	job->minor = snap->sc_snap_minor;
	job->bdfilter = snap->bdfilter;

	wq_insert_work(global_wq(), &job->w);
}

/** Create a block device for the snapshot
 */
static int setup_snapshot_blkdev(struct snapshot_context *snap)
{
	struct gendisk *gd;

	/* Allocate the request queue and use a custom submit_bio_fn.  The original device and COW
	 * device each have their own full request queue, so probably don't need that overhead here.
	 * Newer kernels allocate a disk and a queue at the same time.
	 * Older kernels want you to allocate the disk and queue separately.
	 */
#ifdef HAVE_BLK_ALLOC_DISK
	stcvsm_debug(LOG_SNAPSHOT, "alloc_disk\n");
	snap->sc_snap_gendisk = blk_alloc_disk(NUMA_NO_NODE);
	if (!snap->sc_snap_gendisk) {
		stcvsm_err(LOG_SNAPSHOT, "Failed to allocate gendisk for snapshot\n");
		return -ENOMEM;
	}
	snap->sc_snap_gendisk->minors = 1; /* 1 for "whole disk"; no partitions on a snapshot */
	snap->sc_snap_queue = snap->sc_snap_gendisk->queue;
#else
	stcvsm_debug(LOG_SNAPSHOT, "alloc_disk\n");
	snap->sc_snap_gendisk = alloc_disk(1); /* 1 for "whole disk"; no partitions on a snapshot */
	if (!snap->sc_snap_gendisk) {
		stcvsm_err(LOG_SNAPSHOT, "Failed to allocate gendisk for snapshot\n");
		return -ENOMEM;
	}

#ifdef HAVE_BDOPS_SUBMIT_BIO
	snap->sc_snap_queue = stcvsm_snap_alloc_queue(snap->sc_bdev);
#else
	snap->sc_snap_queue = stcvsm_snap_alloc_queue(snap->sc_bdev, stcvsm_snap_reqfn);
#endif

	snap->sc_snap_gendisk->queue = snap->sc_snap_queue;
#endif // HAVE_BLK_ALLOC_DISK

	if (snap->sc_snap_queue == NULL) {
		del_gendisk(snap->sc_snap_gendisk);
		snap->sc_snap_gendisk = NULL;
		stcvsm_err(LOG_SNAPSHOT, "Cannot allocate request queue\n");
		return -ENOMEM;
	}

	snap->sc_snap_queue->queuedata = snap;

	/* Final snapshot teardown in shutdown_snapshot_blkdev will sync with q,
	 * so no additional refs are taken by queue or workers.
	 */
	gd = snap->sc_snap_gendisk;
	gd->major = stcvsm_major;
	gd->first_minor = snap->sc_snap_minor;
	gd->fops = &snap_ops;
	/* Much like NBD and ramdisk, suppress adding to /proc/partitions. */
	gd->flags |= (GENHD_FL_SUPPRESS_PARTITION_INFO | GENHD_FL_REMOVABLE);
	gd->private_data = snap->bdfilter;
	get_snapshot_name(gd->disk_name, snap->sc_snap_minor);
	stcvsm_info(LOG_SNAPSHOT, "set_capacity %p %llu-%llu\n",
		gd,
		(unsigned long long)snap->sc_sector_end,
		(unsigned long long)snap->sc_sector_start);
	set_capacity(gd, snap->sc_sector_end - snap->sc_sector_start);
	set_disk_ro(gd, 1);  /* FS journals prefer this over -EROFS later */

	return 0;
}

static void remove_snapshot_from_filter(struct snapshot_context *snap)
{
	struct block_device_filter *bdfilter = snap->bdfilter;
	int idx, idx1;

	spin_lock(bdf_lock(bdfilter));
	for (idx = 0; idx < bdfilter->num_snapshots; ++idx) {
		if (bdfilter->snapshots[idx] == snap) {
			--bdfilter->num_snapshots;
			/* Keep the array from getting fragmented */
			for (idx1 = idx; idx1 < bdfilter->num_snapshots; ++idx1)
				bdfilter->snapshots[idx1] = bdfilter->snapshots[idx1 + 1];
			bdfilter->snapshots[idx1] = NULL;
			break;
		}
	}
	spin_unlock(bdf_lock(bdfilter));
}

static int add_snapshot_to_filter(struct block_device_filter *bdfilter, struct snapshot_context *snap)
{
	int i;
	int err = 0;

	spin_lock(bdf_lock(bdfilter));
	if (bdfilter->quiesced) {
		spin_unlock(bdf_lock(bdfilter));
		return -EBUSY;
	}

	for (i = 0; i < bdfilter->num_snapshots; ++i) {
		if (bdfilter->snapshots[i]->sc_bdev->bd_dev == snap->sc_bdev->bd_dev) {
			stcvsm_warning(LOG_SNAPSHOT, "Snapshot of (%pg) already exists on filter\n",
				snap->sc_bdev);
			err = -EBUSY;
			break;
		}
	}
	if (ERROR_INJECT(ERRINJ_SNAPSHOT_ALREADY))
		err = -EBUSY;
	if (err == 0) {
		stcvsm_atomic_inc(&bdfilter->bdf_refs);
		bdfilter->snapshots[bdfilter->num_snapshots] = snap;
		++bdfilter->num_snapshots;
		snap->bdfilter = bdfilter;
	}
	spin_unlock(bdf_lock(bdfilter));
	return err;
}

void put_snapshot(struct snapshot_context *snap)
{
	struct block_device_filter *bdfilter = snap->bdfilter;

	spin_lock(bdf_lock(snap->bdfilter));
	if (stcvsm_atomic_dec_and_test(&snap->sc_refs)) {
		UASSERT(test_bit(SNAPSHOT_SHUTDOWN, &snap->sc_flags), "Shutdown bit not set");
		// Unhook the snapshot from the queue, so any subsequent open requests will fail gracefully
		// TODO: Need locking?
		snap->sc_snap_queue->queuedata = NULL;
		spin_unlock(bdf_lock(snap->bdfilter));
		shutdown_snapshot_blkdev(snap);
		put_filter(bdfilter);
		free_snapshot(snap);
	}
	else
	{
		spin_unlock(bdf_lock(snap->bdfilter));
	}
}

struct snapshot_context *init_snapshot(struct block_device_filter *bdfilter,
		struct block_device *bdev, dev_t cow_dev)
{
	struct snapshot_context *snap;
	int err = 0;

	snap = alloc_snapshot(bdev, cow_dev);
	if (IS_ERR(snap)) {
		err = PTR_ERR(snap);
		stcvsm_err(LOG_SNAPSHOT, "Could not get a snapshot_context (%d)\n", err);
		goto err_alloc;
	}
	// Grabs reference to bdfilter
	err = add_snapshot_to_filter(bdfilter, snap);
	if (err) {
		stcvsm_err(LOG_SNAPSHOT, "Could not add snapshot to filter (%d)\n", err);
		goto err_add_snapshot;
	}
	// Ensure data is filtered before exposing blkdev to user space
	spin_lock(bdf_lock(bdfilter));
	activate_filter(bdfilter);
	spin_unlock(bdf_lock(bdfilter));
	// Expose new snapshot blkdev to user space
	err = get_snapshot_minor(snap);
	if (err) {
		stcvsm_err(LOG_SNAPSHOT, "Could not get a minor (%d)\n", err);
		goto err_get_minor;
	}
	err = setup_snapshot_blkdev(snap);
	if (err)
		goto err_snapshot_blkdev;

	return snap;

err_snapshot_blkdev:
err_get_minor:
	// Drop reference from add_snapshot_to_filter
	put_filter(bdfilter);
	remove_snapshot_from_filter(snap);
err_add_snapshot:
	free_snapshot(snap);
err_alloc:
	return ERR_PTR(err);
}

void stcvsm_shutdown_snapshot(struct snapshot_context *snap)
{
	stcvsm_notice(LOG_SNAPSHOT, "Shutting down %d(%p)\n", atomic_read(&snap->sc_refs), snap);
	/* Mark as shutdown, and remove from list so no new refs can be acquired */
	if (test_and_set_bit(SNAPSHOT_SHUTDOWN, &snap->sc_flags))
		return;
	remove_snapshot_from_filter(snap);

	if (snap->vdiff) {
		if (snap->vdiff->vdiff_orig && !test_bit(VDIFF_RESET, &snap->vdiff->vd_flags))
			rollback_vdiff(snap->vdiff);
		put_vdiff(snap->vdiff);
	}
	cstor_shutdown(snap->cstor);

	/* Release the (possibly) final reference from init_snapshot */
	put_snapshot(snap);
}

#ifdef HAVE_HD_STRUCT_IN_FLIGHT
static void update_in_flight(int cpu, struct hd_struct *hd)
{
/* NOTE: As of 4.15, part_round_stats supports updating the parent device of a
 * partition. Until stcvsm wants to support IO metrics per partition, we can
 * avoid the extra complexity.
 * NOTE: As of 4.15, part_round_stats supports multiqueue devces. Until stcvsm
 * wants to support multiqueue IO, we can avoid that extra complexity too.
 */
	int in_flight = 0;
	unsigned long now = jiffies;

	if (hd->stamp == now)
		return;

	in_flight = atomic_read(&hd->in_flight[0]) + atomic_read(&hd->in_flight[1]);
	if (in_flight > 0) {
		part_stat_add(cpu, hd, time_in_queue,
				in_flight * (now - hd->stamp));
		part_stat_add(cpu, hd, io_ticks, (now - hd->stamp));
	}
	hd->stamp = now;
}
#endif

#ifdef HAVE_HD_STRUCT_IN_FLIGHT
static void diskstat_in(struct hd_struct *hd, int rw)
{
	int cpu = 0;

	cpu = part_stat_lock();
	update_in_flight(cpu, hd);
	atomic_inc(&hd->in_flight[rw]);
	part_stat_unlock();
}
#endif

#ifdef HAVE_HD_STRUCT_IN_FLIGHT
static void diskstat_out(struct hd_struct *hd, unsigned long start_time, unsigned long sectors, int rw)
{
	int cpu = 0;
	unsigned long elapsed = jiffies - start_time;

	cpu = part_stat_lock();
	part_stat_add(cpu, hd, ticks[rw], elapsed);
	part_stat_inc(cpu, hd, ios[rw]);
	part_stat_add(cpu, hd, sectors[rw], sectors);
	update_in_flight(cpu, hd);
	atomic_dec(&hd->in_flight[rw]);
	part_stat_unlock();
}
#endif

/** Waits for the copy step of copy-on-write to finish, before submitting the original write.
 */
static void post_cow_writer(struct wq_work *work)
{
	struct cow_writer_job *job = (struct cow_writer_job *)work;
	struct snapshot_context *snap = job->snap;
	struct block_device_filter *bdfilter = snap->bdfilter;
	int error;
	unsigned int i;
	submit_bio_fn *submit_bio = bdf_get_orig_submit_bio(bdfilter);

	error = cow_wait_fill(snap->cstor, job->cows, job->num_cows);
	if (error) {
		if (!snap->cow_enospc || error != -ENOSPC)
			stcvsm_err(LOG_SNAPSHOT, "error populating cows: %d\n", -error);
		if (!snap->cow_enospc && error == -ENOSPC)
			snap->cow_enospc = 1;
		/* Still allow original write to go through.  Failed cows are now in an error
		 * state so any reads from those locations of the snapshot will fail.
		 */
	}

	for (i = 0; i < job->num_cows; ++i) {
		if (job->cows[i])
			cow_put(snap->cstor, job->cows[i]);
	}

	stcvsm_release_bio(job->orig_bio, submit_bio);
	dec_filtered_writes(bdfilter);

	put_snapshot(snap);
	kfree(job);
}

/** Overlay data from a set of cows onto a completed bio.  Release (pageout) the cows when done.
 */
static void cow_overlay(struct cow_overlay_job *job, int error)
{
	int i;
	struct bio *bio = job->orig_bio;
	struct cow_storage *cstor = job->cstor;
	struct cow_buffer **cows = job->cows;
	int num_cows = job->num_cows;
	int num_readers = 0;

	/* Optimization; no need to map the bio in the common case that all cows were barriers */
	for (i = 0; i < num_cows; ++i)
		num_readers += job->readers[i];

	if (error == 0 && num_cows > num_readers) {
		bio_iter_t iter;
		bio_iter_bvec_t bv;
		sector_t sector; /* current sector */
		char *cowbuf = NULL;
		const unsigned int sector_size = 512;  /* bio "sectors" are 512 bytes, by definition */
		unsigned int cow_offset, page_index, page_offset;

		/* Loop over the bio_vecs.  Take care to ensure that kmap_atomic are strictly
		 * nested.
		 */
		sector = bio_sector(bio);
		i = -1;
		bio_for_each_segment(bv, bio, iter) {
			unsigned int bv_offset;
			char *bv_buf = stc_kmap_atomic_bvec(bv);

			for (bv_offset = 0; bv_offset < bio_iter_len(bio, iter); bv_offset += sector_size, ++sector) {
				sector_t s = sector;

				cow_offset = do_div(s, cstor->sectors_per_cow) << 9;
				if (cow_offset == 0 || i == -1)
					i++;
				page_offset = cow_offset & (PAGE_SIZE-1);
				page_index = cow_offset >> PAGE_SHIFT;
				if (page_offset == 0 || !cowbuf) {
					if (cowbuf) {
						stc_kunmap_atomic(cowbuf);
						cowbuf = NULL;
					}
					/* Cows that acted as a read lock do not have pages */
					if (!job->readers[i])
						cowbuf = stc_kmap_atomic(cows[i]->pages[page_index]);
				}
				if (cowbuf)
					memcpy(bv_buf+bv_offset, cowbuf+page_offset, sector_size);
			}

			if (cowbuf) {
				stc_kunmap_atomic(cowbuf);
				cowbuf = NULL;
			}

			stc_kunmap_atomic(bv_buf);
		}
	}

	for (i = 0; i < num_cows; ++i)
		cow_put(cstor, cows[i]);
}

/** Waits for cows to finish paging in, then overlay cows onto the completed bio from the original
 * disk.
 */
static void wait_cow_and_overlay(struct wq_work *w)
{
	struct cow_overlay_job *job = (struct cow_overlay_job *)w;
	int error = cow_wait_pagein(job->cstor, job->cows, job->num_cows);
#ifdef HAVE_HD_STRUCT_IN_FLIGHT
	int rw = bio_data_dir(job->orig_bio);
	unsigned long sectors = bio_sectors(job->orig_bio);
#endif

	cow_overlay(job, error);
	stc_bio_endio(job->orig_bio, error);
#ifdef HAVE_HD_STRUCT_IN_FLIGHT
	diskstat_out(job->hd, job->start_time, sectors, rw);
#endif
	kfree(job);
}

/** Called when a bio from original disk completes.  Queues everything to be completed in sleepable
 * context.
 */
static void cow_overlay_queue(struct cow_overlay_job *job, int error)
{
	if (!error) {
		job->w.fn = wait_cow_and_overlay;
		wq_insert_work(&job->cstor->cow_wq, &job->w);
	} else {
		stc_bio_endio(job->orig_bio, error);
#ifdef HAVE_HD_STRUCT_IN_FLIGHT
		{
			int rw = bio_data_dir(job->orig_bio);
			unsigned long sectors = bio_sectors(job->orig_bio);
			diskstat_out(job->hd, job->start_time, sectors, rw);
		}
#endif
		kfree(job);
	}
}

static BIO_END_IO_PROTO(cow_overlay_finish_bio, bio)
{
	unsigned int i;
	struct cow_overlay_job *job = bio->bi_private;

	for (i = 0; i < job->num_cows; ++i) {
		if (job->readers[i])
			cow_reader_unlock(job->cstor, &job->cows[i], 1);
	}

	cow_overlay_queue(job, BIO_ERROR(bio));
	bio_put(bio);
}

/** Redirects a read bio to the original disk, and prepares to overlay the cows.
 */
static int bio_redirect_and_overlay(struct cow_overlay_job *job, struct block_device *bdev)
{
	struct bio *bio;

	bio = stc_bio_clone(job->orig_bio, GFP_NOIO);
	if (!bio)
		return -ENOMEM;
	bio->bi_end_io = cow_overlay_finish_bio;
	stc_bio_assign_disk(bio, bdev);
	bio->bi_private = job;
	stc_submit_bio(bio);
	return 0;
}

static BIO_END_IO_PROTO(cow_fill_endio, bio)
{
	struct cow_fill_job *job = bio->bi_private;

	if (BIO_ERROR(bio))
		job->error = BIO_ERROR(bio);
	if (atomic_dec_and_test(&job->num_bios))
		cow_finish_fill(job->cstor, &job->cow, 1, job->error);
	bio_put(bio);
}

/** Filters writes to the original disk, initiating copy-on-write if necessary.
 *
 * @return 1 if the original write has been filtered due to COW, else 0
 */
int orig_write_filter(struct request_queue *q, struct bio *bio, struct block_device_filter *bdfilter,
		struct snapshot_context *snap)
{
	int retval = 0;
	sector_t cow_start, cow_end;
	struct cow_storage *cstor = snap->cstor;
	int i, num_cows = 0, max_cows;
	struct cow_buffer **cows;
	struct cow_fill_job *cow_fill_jobs;
	struct cow_writer_job *job;

	/* Calculate cows relative to the start of the snap->sc_bdev struct */
	cow_start = bio_sector(bio) - snap->sc_sector_start;
	cow_end = cow_start + bio_sectors(bio);
	do_div(cow_start, cstor->sectors_per_cow);
	if (do_div(cow_end, cstor->sectors_per_cow))
		++cow_end;
	max_cows = cow_end - cow_start;

	/* Combine several allocations in one, to reduce possible failures. */
	{
		size_t job_size = sizeof(struct cow_writer_job);
		size_t cows_size = sizeof(struct cow_buffer *) * max_cows;
		size_t fill_job_size = sizeof(struct cow_fill_job) * max_cows;

		if (ERROR_INJECT(ERRINJ_ALLOC))
			job = NULL;
		else
			job = kzalloc(job_size + cows_size + fill_job_size, GFP_NOFS);
		if (!job) {
			retval = -ENOMEM;
			goto snap_error;
		}
		cows = (struct cow_buffer **)(((char *)job) + job_size);
		cow_fill_jobs = (struct cow_fill_job *)(((char *)cows) + cows_size);
	}

	/* Claim some cows to populate. */
	for (i = 0; i < max_cows; ++i) {
		struct cow_buffer *cow = cow_get(cstor, (u32)cow_start+i);

		if (!cow) {
			retval = -ENOMEM;
			goto snap_error;
		}
		retval = cow_start_fill(cstor, cow);
		if (retval == 0) {
			num_cows++;
		} else if (retval == -EBUSY) {
			/* Already busy filling -- not my job to kick off the fill but must still
			 * wait for it to complete.
			 */
			cow_fill_jobs[i].error = -EBUSY;
			num_cows++;
		} else if (retval == 1) {
			/* This cow is already done or failed or freespace; don't fill it again */
			cow_put(cstor, cow);
			cow = NULL;
		} else {
			stcvsm_err(LOG_COWS, "cow %u failed: %d\n", (unsigned int)cow->index, -retval);
			/* Allow the write to proceed, but attempting to read this block in the
			 * snapshot will fail.
			 */
			cow_put(cstor, cow);
			cow = NULL;
		}
		cows[i] = cow;
	}

	if (num_cows == 0) {
		kfree(job);
		return 0;
	} else {
		struct blk_plug plug;

		blk_start_plug(&plug);

		for (i = 0; i < max_cows; ++i) {
			struct cow_fill_job *fill_job = &cow_fill_jobs[i];

			if (cows[i] && fill_job->error != -EBUSY) {
				struct bio *bios;

				fill_job->cstor = cstor;
				fill_job->cow = cows[i];
				retval = cow_pack_bios(fill_job->cstor, cows[i], cows[i]->index, snap->sc_bdev, &bios);
				if (retval > 0) {
					struct bio *bio;

					atomic_set(&fill_job->num_bios, retval);
					retval = 0;
					for (bio = bios; bio; bio = bios) {
						bios = bio->bi_next;
						bio->bi_next = NULL;
						/* Reads are realtime class, because they are blocking writes */
						bio_set_prio(bio, IOPRIO_PRIO_VALUE(IOPRIO_CLASS_RT, 0));
						bio->bi_private = fill_job;
						stc_bio_assign_disk(bio, snap->sc_bdev);
						bio->bi_end_io = cow_fill_endio;
						stc_submit_bio(bio);
					}
				} else {
					cow_finish_fill(cstor, &cows[i], 1, retval);
				}
			}
		}
		blk_finish_plug(&plug);

		/* Now wait for the reads into the COWs to complete before issuing the write. */
		stcvsm_atomic_inc(&snap->sc_refs);
		job->w.fn = post_cow_writer;
		job->cows = cows;
		job->num_cows = max_cows;
		job->orig_bio = bio;
		job->snap = snap;
		wq_insert_work(&cstor->cow_wq, &job->w);
		return 1;
	}

snap_error:
	stcvsm_err(LOG_SNAPSHOT, "Shutting down snapshot due to error %d\n", -retval);
	kfree(job);

	/* The snapshot is no longer in a known state; mark as ERROR to prevent new IO */
	set_bit(SNAPSHOT_ERROR, &snap->sc_flags);

	return 0;
}

int activate_snapshots(struct list_head *snap_list)
{
	struct snapshot_context *snap;
	int err = 0;

	list_for_each_entry(snap, snap_list, set_peer_list) {
		stcvsm_notice(LOG_SNAPSHOT, "Activating snapshot in list: %s\n", snap->sc_snap_gendisk->disk_name);
		/* Freeze the block device asked for, so we can get a consistent snapshot. */
		err = stcvsm_freeze_bdev(snap->sc_bdev, &snap->sc_frozen_sb);
		if (err < 0) {
			stcvsm_err(LOG_SNAPSHOT, "Failed to freeze block device (%d)\n", err);
			break;
		}
	}
	if (err) {
		stcvsm_warning(LOG_SNAPSHOT,
			"Freezing all SuperBlocks failed (%d), thawing all devices in snapshot set\n", err);
		list_for_each_entry(snap, snap_list, set_peer_list) {
			if (snap->sc_frozen_sb) {
				stcvsm_thaw_bdev(snap->sc_bdev, snap->sc_frozen_sb);
				snap->sc_frozen_sb = NULL;
			}
		}
		return err;
	}
	list_for_each_entry(snap, snap_list, set_peer_list) {
		struct vdiff_context *vdiff;

		set_bit(SNAPSHOT_SNAPPED, &snap->sc_flags);

		stcvsm_notice(LOG_SNAPSHOT, "add_disk %s\n", snap->sc_snap_gendisk->disk_name);

		vdiff = get_vdiff_bdev(snap->bdfilter, snap->sc_bdev);
		if (vdiff) {
			snap->vdiff = copy_reset_vdiff(snap->bdfilter, vdiff,
					MKDEV(stcvsm_major, snap->sc_snap_minor));
			if (IS_ERR(snap->vdiff)) {
				err = PTR_ERR(snap->vdiff);
				stcvsm_err(LOG_SNAPSHOT, "Failed to copy_reset_vdiff (%d)\n", err);
				snap->vdiff = NULL;
			}
			put_vdiff(vdiff);
		} else {
			stcvsm_info(LOG_SNAPSHOT, "Does not have vdiff!\n");
		}

		if (snap->sc_frozen_sb) {
			stcvsm_thaw_bdev(snap->sc_bdev, snap->sc_frozen_sb);
			snap->sc_frozen_sb = NULL;
		}
	}
	list_for_each_entry(snap, snap_list, set_peer_list) {
		set_bit(SNAPSHOT_INIT, &snap->sc_flags);
		add_disk(snap->sc_snap_gendisk);
	}
	return err;
}

/** Start the process of reading from the original volume (via a cloned bio), and the cow, and
 * merging the results.
 *
 * @return 0 if the read has been started, non-zero on error
 */
static int stcvsm_snapshot_read(struct snapshot_context *snap, unsigned long start_time, struct bio *bio)
{
	int retval;
	sector_t cow_start, cow_end;
	struct cow_storage *cstor = snap->cstor;
	int i, num_cows, num_readers = 0;
	struct cow_buffer **cows;
	struct cow_overlay_job *job;

	/* Convert from sectors to cows */
	cow_start = bio_sector(bio);
	cow_end = cow_start + bio_sectors(bio);
	do_div(cow_start, cstor->sectors_per_cow);
	if (do_div(cow_end, cstor->sectors_per_cow))
		++cow_end;
	num_cows = cow_end - cow_start;

	/* Combine several allocations in one, to reduce possible failures. */
	{
		size_t job_size = sizeof(struct cow_overlay_job);
		size_t cows_size = sizeof(struct cow_buffer *) * num_cows;

		if (ERROR_INJECT(ERRINJ_ALLOC))
			job = NULL;
		else
			job = kzalloc(job_size + cows_size + num_cows, GFP_NOFS);
		if (!job) {
			retval = -ENOMEM;
			goto out;
		}
		cows = (struct cow_buffer **)(((char *)job) + job_size);
		job->readers = (u8 *)(((char *)cows) + cows_size);
	}

	/* Find all cows that overlap the read request, and start pulling them into memory. */
	for (i = 0; i < num_cows; ++i) {
		int r;
		struct cow_buffer *cow;

		cow = cow_get(cstor, (u32)cow_start+i);
		if (!cow) {
			retval = -ENOMEM;
			num_cows = i;
			goto pageout;
		}

		job->readers[i] = 0;
		/* Do not care about throttling memory usage in cow_start_pagein when reading COWs
		 * for the snapshot.  The COW read bio and the redirected read bio are submitted to
		 * (potentially) separate request queues in parallel, so worst case is 2x memory
		 * usage.
		 */
		r = cow_start_pagein(cstor, cow);
		if (r < 0) {
			retval = r;
			num_cows = i;
			goto pageout;
		} else if (r > 0) {
			job->readers[i] = 1;
			num_readers++;
		}
		cows[i] = cow;
	}

	job->cstor = cstor;
	job->cows = cows;
	job->num_cows = num_cows;
	job->orig_bio = bio;
#ifdef HAVE_HD_STRUCT_IN_FLIGHT
	job->hd = &snap->sc_snap_gendisk->part0;
#endif
	job->start_time = start_time;

	if (num_readers == 0) {
		/* All cows contain copy-on-write data; no need to read from original device.
		 * Pretend the no-op read from the original device succeeded, and overlay the cows.
		 */
		cow_overlay_queue(job, 0);
	} else {
		retval = bio_redirect_and_overlay(job, snap->sc_bdev);
		if (retval)
			goto pageout;
	}

	return 0;

pageout:
	for (i = 0; i < num_cows; ++i) {
		if (job->readers[i])
			cow_reader_unlock(cstor, &cows[i], 1);
		cow_put(cstor, cows[i]);
	}
	kfree(job);
	if (!snap->cow_enospc || retval != -ENOSPC)
		stcvsm_err(LOG_SNAPSHOT, "cow %u failed: %d\n", (unsigned int)cow_start+i, -retval);
	if (!snap->cow_enospc && retval == -ENOSPC)
		snap->cow_enospc = 1;
out:
	return retval;
}

/** The snapshot device's request function.  We don't support writeable snapshots (yet), so write
 * requests fail.  Read requests are shipped off to stcvsm_snapshot_read for processing.
 */
void __submit_bio(struct request_queue *q, struct bio *bio)
{
	// TODO: Perhaps we need to lock the queue, grab snap and a reference on it.
	//     : Perhaps, we only get submit_bio calls between open and release, and this is safe.
	struct snapshot_context *snap = q->queuedata;
	int error = 0;
	int do_endio = 1;
	int rw = bio_data_dir(bio);
	unsigned long start_time = jiffies;

	stcvsm_debug(LOG_IO, "[%llu] [%s] sectors %u\n",
		(unsigned long long)bio_sector(bio),
		rw == WRITE ? "write" : "read",
		bio_sectors(bio));

	if (!bio_has_data(bio)) {
		/* BIOs that have no data are control messages (FUA, discard, ...)
		 * which aren't relevant for the snapshot.
		 */
		trace_snap_ctl_reqfn_io(stcvsm_major, snap->sc_snap_minor, 0, 0);
	} else if (rw == WRITE) {
		error = -EROFS;
		trace_snap_write_reqfn_io(stcvsm_major, snap->sc_snap_minor, bio_sector(bio), bio_sectors(bio));
	} else {
		if (test_bit(SNAPSHOT_SHUTDOWN, &snap->sc_flags)) {
			/* A shutdown call from userspace is authoritative.  Any trailing IO is
			 * rogue, and gets zeroes.  Specifically do not want nasty IO errors
			 * logged for IO we did not initiate.
			 */
			stcvsm_debug(LOG_IO, "[%llu] Snapshot shutdown, so zero buffer\n",
					(unsigned long long)bio_sector(bio));
			if (bio_has_data(bio) && rw == READ) {
				bio_iter_bvec_t bv;
				bio_iter_t iter;

				bio_for_each_segment(bv, bio, iter) {
					char *bv_buf = stc_kmap_atomic_bvec(bv);
					memset(bv_buf, 0, bio_iter_len(bio, iter));
					stc_kunmap_atomic(bv_buf);
				}
			}
		} else if (test_bit(SNAPSHOT_ERROR, &snap->sc_flags)) {
			/* Failed snapshots are a true problem, and so we do want IO errors. */
			// TODO: Move this to info or worse
			stcvsm_warning(LOG_SNAPSHOT, "Snapshot in ERROR state");
			error = -EIO;
		} else {
#ifdef HAVE_HD_STRUCT_IN_FLIGHT
			diskstat_in(&snap->sc_snap_gendisk->part0, rw);
#endif

			trace_snap_read_reqfn_io(stcvsm_major, snap->sc_snap_minor, bio_sector(bio), bio_sectors(bio));
			error = stcvsm_snapshot_read(snap, start_time, bio);
			if (error == 0)
				do_endio = 0;
#ifdef HAVE_HD_STRUCT_IN_FLIGHT
			else
				diskstat_out(&snap->sc_snap_gendisk->part0, start_time, 0, rw);
#endif
		}
	}
	if (do_endio)
		stc_bio_endio(bio, error);
}

#ifdef HAVE_BDOPS_SUBMIT_BIO
static MAKE_REQUEST_RETTYPE snap_submit_bio(struct bio *bio)
{
	__submit_bio(bio_request_queue(bio), bio);
#ifdef MAKE_REQUEST_RETVAL
	return MAKE_REQUEST_RETVAL;
#endif
}
#else
static MAKE_REQUEST_RETTYPE stcvsm_snap_reqfn(struct request_queue *q, struct bio *bio)
{
	__submit_bio(q, bio);
#ifdef MAKE_REQUEST_RETVAL
	return MAKE_REQUEST_RETVAL;
#endif
}
#endif


#ifdef WITH_EXT2BITMAP
/** Allocate and populate a free-space bitmap buffer with the bitmap from the filesystem.
 */
static int stcvsm_fs_bitmap(struct snapshot_context *snap, struct super_block *sb)
{
	int err = 0;

	if (!sb)
		return -ENOENT;
	/* Create a bitmap for all of the sectors for the snapshotted device. */
	snap->freespc_bitmap = kmalloc_array(BITS_TO_LONGS(snap->bits_in_map), sizeof(*snap->freespc_bitmap), GFP_KERNEL);
	if (!snap->freespc_bitmap)
		return -ENOMEM;
	bitmap_zero(snap->freespc_bitmap, snap->bits_in_map);
	if (strncmp(sb->s_type->name, "ext", 3) == 0) {
		struct block_device *snap_bdev;

		stcvsm_debug(LOG_SNAPSHOT, "Trying out %s bitmap\n", sb->s_type->name);
		snap_bdev = stcvsm_open_blkdev(MKDEV(stcvsm_major, snap->sc_snap_minor), FMODE_READ, NULL);
		if (IS_ERR(snap_bdev)) {
			err = PTR_ERR(snap_bdev);
		} else {
			/* TODO: We should probably be passing down cstor here */
			ext2_create_bitmap(snap_bdev, snap->freespc_bitmap, snap->bits_in_map,
					snap->block_size);
			stcvsm_close_blkdev(snap_bdev, FMODE_READ);
		}
	} else {
		err = -EINVAL;
	}
	if (err) {
		kfree(snap->freespc_bitmap);
		snap->freespc_bitmap = NULL;
	}
	return err;
}
#endif

const char *snapshot_state_to_str(unsigned long state)
{
	if (!(test_bit(SNAPSHOT_INIT, &state)))
		return "uninit";
	else if (test_bit(SNAPSHOT_SHUTDOWN, &state))
		return "shutdown";
	else if (test_bit(SNAPSHOT_ERROR, &state))
		return "error";
	else
		return "active";
}
