/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#ifndef STCSNAPDEF_H
#define STCSNAPDEF_H
// Constants

// Snap mode bitwise flag values - can be ORed together

// Snapshot will be read-only
#define VSNAP_SNAP_READ_ONLY                1
// Opens "with backup intent" will be reparsed to the snapshot
#define VSNAP_SNAP_REPARSE_ON_BACKUP        2
// Forces the engine to use the usual B-tree
// Compressed tables will be used otherwise
// VSNAP_SNAP_USE_THIRD_BITMAP is ignored (as if not set) with compressed tables
#define VSNAP_SNAP_DONT_COMPRESS_TABLES     4
// Forces the engine to use usual bitmaps
// Compressed bitmaps will be used otherwise
// VSNAP_SNAP_USE_THIRD_BITMAP is ignored (as if not set) with compressed bitmaps
#define VSNAP_SNAP_DONT_COMPRESS_BITMAPS    8
// Use the third additional bitmap in the engine to speed up some things
// Rarely used, unnecessary (and thus ignored) with compressed tables or bitmaps, and thus
//  is in effect on for B-tree and usual bitmaps
#define VSNAP_SNAP_USE_THIRD_BITMAP         16
#ifndef NO_HR
// Snaps for future possible hot rollback
#define VSNAP_SNAP_FOR_ROLLBACK             32
#endif
#ifndef NO_VDIFF
// VDIFF is not touched by snap/unsnap activity at all if this flag is set
// Otherwise, snap/unsnap will cause VDIFF updates (migrating the diff table to snapshot
//  and so on)
#define VSNAP_SNAP_RETAIN_VDIFF             64
#endif

#ifdef RWE

// Forces the engine to use redundant write elimination (read before write, then
//  data compare)
// Only rarely useful, and only disk-space-wise, a noticeable performance killer
#define VSNAP_SNAP_USE_READ_BEFORE_WRITE    0x10000

#endif

// Phase numbers to determine where VsnapSnapTheSet have failed exactly

#define VSNAP_SET_PHASE_END_PREPARE         1
#define VSNAP_SET_PHASE_PRE_COMMIT          2
#define VSNAP_SET_PHASE_COMMIT              3
#define VSNAP_SET_PHASE_POST_COMMIT         4

// Verb codes for VsnapSnapshotDoVerbByHandle
// Results for VsnapGetStats
typedef struct _VSNAP_QUERY_STATS_EX_OUTPUT_BUFFER
{
    // Total blocks (clusters/allocation units) for the volume
    // This includes the FATs and root directory for FAT, which are divided to "virtual blocks"
    uint32_t NumberOfBlocks;
    // Number of blocks available for use at the time of snap
    uint32_t NumberOfAvailableBlocksOnSnap;
    // Number of blocks available for use by FS and VSNAP
    uint32_t NumberOfAvailableBlocks;
    // Number of blocks allocated to VSNAP's stored data
    uint32_t NumberOfVsnapBlocks;
    // Number of writes passed through the VSNAP engine since snapshot was established
    uint32_t NumberOfVsnapWrites;
    // Number of reads passed through the VSNAP engine since snapshot was established
    uint32_t NumberOfVsnapReads;
    // Number of times a store block was written, not including the redirections
    uint32_t NumberOfStoreWrites;
    // Number of times a store block was read
    uint32_t NumberOfStoreReads;
    // Number of collisions resolved by the VSNAP engine
    // Collision occurs when the FS wants to write something to the block busy with VSNAP
    //  data
    uint32_t NumberOfResolvedCollisions;
    // The time when snapshot was imposed in FILETIME format
    LARGE_INTEGER SnapTime;
    // TRUE if the snapshot is "cold"
    uint8_t IsColdSnap;
} VSNAP_QUERY_STATS_EX_OUTPUT_BUFFER, *PVSNAP_QUERY_STATS_EX_OUTPUT_BUFFER;

// Results for VsnapGetProductInfo
typedef struct _VSNAP_PRODUCT_INFO
{
    // Total length of the structure in bytes including the strings at tail
    // The length of the buffer required to return the whole information
    uint32_t TotalLength;
    // Length in bytes of the ProductName string (including the terminating zero)
    //  at the tail of this structure
    uint32_t ProductNameLength;
    // Length in bytes of the VersionDescription string (including the terminating zero)
    //  at the tail of this structure
    // VersionDescription immediately follows the ProductName
    uint32_t VersionDescriptionLength;
    // Major version number
    uint32_t MajorVersion;
    // Minor version number
    uint32_t MinorVersion;
    // Days in the product expiration period
    // This is the overall expiration period as hard-coded to the binary, and not the
    //  not-yet-expired part of it
    uint32_t ExpirationPeriod;
    // Flags which denote the build options linked in the binary
    // See VSNAP_PRODUCT_INFO_FLAG_xxx above
    uint32_t Flags;
    // ProductName zero-terminated Unicode string follows
    // VersionDescription zero-terminated Unicode string follows after ProductName
} VSNAP_PRODUCT_INFO, *PVSNAP_PRODUCT_INFO;

// VDIFF information queried by VsnapDiffGetInfo
typedef struct _VSNAP_VDIFF_INFO
{
    // TRUE if volume has VDIFF active
    uint8_t VDiffActive;
    // Percentage of dirty (written since last backup) sectors on the volume tracked by VDIFF
    // 100 if VDIFF is not active
    uint32_t VDiffPercentsDirty;
    // VDIFF GUID
    // New GUID is invented on each VDIFF startup
    GUID VDiffGuid;
    // VDIFF generation count
    uint32_t VDiffResetCount;
} VSNAP_VDIFF_INFO, *PVSNAP_VDIFF_INFO;

// VDIFF information queried by VsnapDiffGetInfo2
typedef struct _VSNAP_VDIFF_INFO_2
{
    // TRUE if volume has VDIFF active
    uint8_t VDiffActive;
    // Bytes per sector
    uint32_t SectorSize;
    // Dirty volume sectors
    uint64_t DirtySectors;
    // Total volume sectors
    uint64_t TotalSectors;
    // VDIFF GUID
    // New GUID is invented on each VDIFF startup
    GUID VDiffGuid;
    // VDIFF file GUID
    // On each VDIFF startup and major reset, the file GUID is set to all zeroes
    // Can also be explicitly set by VsnapDiffSetFileGuidByHandle or by VsnapDiffReanimate
    // Persistent across boots together with the rest of the VDIFF data until
    //  VDIFF will be turned off or a crash
    GUID VDiffFileGuid;
    // VDIFF generation count
    uint32_t VDiffResetCount;
} VSNAP_VDIFF_INFO_2, *PVSNAP_VDIFF_INFO_2;

typedef struct _VSNAP_IOCTL_INFO
{
    int major;
    int minor;
    VSNAP_VDIFF_INFO_2 Info;
} VSNAP_IOCTL_INFO, *PVSNAP_IOCTL_INFO;

#define VSNAP_DEVICE_NAME "/dev/stcvsmctl"
#endif // STCSNAPDEF_H
