/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include "stcvsm_config.h"

#include <linux/compat.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <linux/string.h>
#ifdef HAVE_INCLUDE_VERMAGIC
#include <linux/utsname.h>
#else
// Kernel 5.8 introduces guards around using VERMAGIC values, so stop using them.
#include <linux/vermagic.h>
#endif

#include "debug.h"
#include "filter.h"
#include "module.h"
#include "snapshot.h"
#include "snapshot_set.h"
#include "stcvsm_version.h"
#include "stcvsmctl.h"
#include "vdiff.h"

#include "stcsnap_p.h"
#include "stcsnapdef_p.h"

#define STCVSMCTL_DEVICE_NAME "stcvsmctl"

static int get_product_info(VSNAP_PRODUCT_INFO __user *info)
{
	int retval;
	u32 avail, product_len, version_len, len;
	const char *product_name = "stcvsm";
	VSNAP_PRODUCT_INFO *buf;
	char *strings;
#ifdef HAVE_INCLUDE_VERMAGIC
	char *this_kernel_version;
#else
	const char* this_kernel_version;
#endif

	retval = get_user(avail, &info->TotalLength);
	if (retval)
		return retval;

	product_len = strlen(product_name) + 1;
#ifdef HAVE_INCLUDE_VERMAGIC
	this_kernel_version = utsname()->release;
#else
	this_kernel_version = VERMAGIC_STRING;
#endif
	version_len = 8 + sizeof(STCVSM_VERSION) + 9 + strlen(this_kernel_version) + 1;
	len = sizeof(VSNAP_PRODUCT_INFO) + product_len + version_len;
	buf = kmalloc(len, GFP_KERNEL);
	if (buf == NULL)
		return -ENOMEM;
	buf->TotalLength = len;
	buf->ProductNameLength = product_len;
	buf->VersionDescriptionLength = version_len;
	buf->MajorVersion = STCVSM_MAJOR;
	buf->MinorVersion = STCVSM_MINOR;
	buf->ExpirationPeriod = (u32)-1;
	buf->Flags = 0;
	strings = ((char *)buf) + sizeof(VSNAP_PRODUCT_INFO);
	memcpy(strings, product_name, product_len);
	strings += product_len;
	memcpy(strings, "version=", 8);
	strings += 8;
	memcpy(strings, STCVSM_VERSION, sizeof(STCVSM_VERSION));
	strings += sizeof(STCVSM_VERSION);
	memcpy(strings, "vermagic=", 9);
	strings += 9;
	memcpy(strings, this_kernel_version, strlen(this_kernel_version) + 1);

	retval = copy_to_user(info, buf, avail < len ? avail : len);
	kfree(buf);

	return retval ? -EFAULT : 0;
}

int stcvsm_ioctl_snapshot(dev_t dev, unsigned int cmd, void *param)
{
	int retval = 0;
	struct block_device *bdev, *filter_bdev;
	struct block_device_filter *bdfilter;

	bdev = stcvsm_open_blkdev(dev, FMODE_READ, NULL);
	if (IS_ERR(bdev))
		return PTR_ERR(bdev);

	if (MAJOR(dev) == stcvsm_major)
		filter_bdev = get_snapshot_filter_bdev(bdev);
	else
		filter_bdev = bdev;

	bdfilter = get_filter(filter_bdev, cmd == IOCTL_SNAP_VOLUME_EX);
	if (bdfilter && !IS_ERR(bdfilter)) {
		struct snapshot_context *snapshot;

		if (cmd == IOCTL_SNAP_VOLUME_EX) {
			VSNAP_SNAP_INPUT_BUFFER_EX *buf = (VSNAP_SNAP_INPUT_BUFFER_EX *)param;

			dev_t cow_dev = MKDEV(buf->cow_major, buf->cow_minor);

			stcvsm_debug(LOG_FILTER, "Creating snapshot\n");
			/* Adds refcount to bdfilter */
			snapshot = init_snapshot(bdfilter, bdev, cow_dev);
			if (IS_ERR(snapshot)) {
				retval = PTR_ERR(snapshot);
				stcvsm_warning(LOG_FILTER, "Failed to create snapshot (%d)\n", retval);
			} else {
				LIST_HEAD(snap_list);

				list_add(&snapshot->set_peer_list, &snap_list);
				retval = activate_snapshots(&snap_list);
				if (retval) {
					stcvsm_warning(LOG_FILTER, "Failed to activate snapshot (%d)\n", retval);
					stcvsm_shutdown_snapshot(snapshot);
				}
			}
		} else {
			spin_lock(bdf_lock(bdfilter));
			snapshot = get_snapshot_bdev(bdfilter, bdev);
			spin_unlock(bdf_lock(bdfilter));

			if (snapshot == NULL) {
				retval = -ENOENT;
			} else {
				switch (cmd) {
				case IOCTL_UNSNAP_VOLUME:
				{
					u8 null_guid[16] = {0};

					if (memcmp(snapshot->set_guid, null_guid, 16) == 0)
						stcvsm_shutdown_snapshot(snapshot);
					else
						set_unsnap(snapshot->set_guid);
					break;
				}
				case IOCTL_GET_SNAPSHOT_NAME_EX:
				{
					if (test_bit(SNAPSHOT_SHUTDOWN, &snapshot->sc_flags)) {
						retval = -EINVAL;
					} else {
						SNAPSHOT_NAME_BUF *name = (SNAPSHOT_NAME_BUF *)param;

						char *disk_name = snapshot->sc_snap_gendisk->disk_name;
						size_t len = strlen(disk_name) + 1;

						name->path_size = len;
						strncpy(name->path, disk_name, sizeof(name->path));
						if (len > sizeof(name->path))
							retval = -ENOBUFS;
					}
					break;
				}
				default:
					retval = -EOPNOTSUPP;
					break;
				}
				put_snapshot(snapshot);
			}
		}
		put_filter(bdfilter);
	} else {
		retval = bdfilter ? PTR_ERR(bdfilter) : -ENOENT;
	}
	stcvsm_close_blkdev(bdev, FMODE_READ);
	return retval;
}

/* Similar to stcvsm_ioctl_vdiff, but can be called on a snapshot also.
 */
static int vdiff_ioctl_on_snap(dev_t dev, unsigned int cmd, void *param)
{
	int retval;
	struct block_device *bdev, *filter_bdev;
	struct block_device_filter *bdfilter;

	bdev = stcvsm_open_blkdev(dev, FMODE_READ, NULL);
	if (IS_ERR(bdev))
		return PTR_ERR(bdev);

	if (MAJOR(dev) == stcvsm_major)
		filter_bdev = get_snapshot_filter_bdev(bdev);
	else
		filter_bdev = bdev;

	bdfilter = get_filter(filter_bdev, false);
	if (!IS_ERR(bdfilter)) {
		retval = _stcvsm_ioctl_vdiff(bdfilter, bdev, cmd, param);
		if (bdfilter)
			put_filter(bdfilter);
	} else {
		retval = PTR_ERR(bdfilter);
	}
	stcvsm_close_blkdev(bdev, FMODE_READ);
	return retval;
}

static long stcvsmctl_ioctl(struct file *filep, unsigned int cmd, unsigned long arg)
{
	void __user *param = (void __user *)arg;  /* Most/all cmds point to structs in userspace */
	long retval;

	switch (cmd) {
	case IOCTL_VSNAP_GET_PRODUCT_INFO:
	{
		retval = get_product_info(param);
		break;
	}
	case IOCTL_VSNAP_VDIFF_START:
	{
		VSNAP_IOCTL_GUID buf;

		if (copy_from_user(&buf, param, sizeof(buf))) {
			retval = -EFAULT;
		} else {
			stcvsm_notice(LOG_CTL, "IOCTL_VSNAP_VDIFF_START: (%d,%d)\n", buf.major, buf.minor);
			if (buf.major == stcvsm_major)
				retval = -EINVAL;
			else
				retval = stcvsm_ioctl_vdiff(MKDEV(buf.major, buf.minor), cmd, &buf);
		}
		break;
	}
	case IOCTL_VSNAP_VDIFF_STOP:
	{
		VSNAP_IOCTL_GUID buf;

		if (copy_from_user(&buf, param, sizeof(buf))) {
			retval = -EFAULT;
		} else {
			stcvsm_notice(LOG_CTL, "IOCTL_VSNAP_VDIFF_STOP: (%d,%d)\n", buf.major, buf.minor);
			if (buf.major == stcvsm_major)
				retval = -EINVAL;
			else
				retval = stcvsm_ioctl_vdiff(MKDEV(buf.major, buf.minor), cmd, &buf);
		}
		break;
	}
	case IOCTL_VSNAP_VDIFF_STATS_2:  /* Deprecated; incorrect size bits */
	case IOCTL_VSNAP_VDIFF_STATS_3:
	{
		VSNAP_IOCTL_INFO buf;

		if (copy_from_user(&buf, param, sizeof(buf))) {
			retval = -EFAULT;
		} else {
			stcvsm_notice(LOG_CTL, "IOCTL_VSNAP_VDIFF_STATS_2: (%d,%d)\n", buf.major, buf.minor);
			retval = vdiff_ioctl_on_snap(MKDEV(buf.major, buf.minor), IOCTL_VSNAP_VDIFF_STATS_2, &buf);
			if (retval == 0 && copy_to_user(param, &buf, sizeof(buf)))
				retval = -EFAULT;
		}
		break;
	}
	case IOCTL_VSNAP_VDIFF_RESET:
	{
		VSNAP_IOCTL_REANIMATE_BUFF buf;

		if (copy_from_user(&buf, param, sizeof(buf))) {
			retval = -EFAULT;
		} else {
			stcvsm_notice(LOG_CTL, "IOCTL_VSNAP_VDIFF_RESET: (%d,%d)\n", buf.major, buf.minor);
			retval = vdiff_ioctl_on_snap(MKDEV(buf.major, buf.minor), cmd, &buf);
		}
		break;
	}
	case IOCTL_VSNAP_VDIFF_SET_FILE_GUID:
	{
		VSNAP_IOCTL_GUID buf;

		if (copy_from_user(&buf, param, sizeof(buf))) {
			retval = -EFAULT;
		} else {
			stcvsm_notice(LOG_CTL, "IOCTL_VSNAP_VDIFF_SET_FILE_GUID: (%d,%d)\n", buf.major, buf.minor);
			retval = vdiff_ioctl_on_snap(MKDEV(buf.major, buf.minor), cmd, &buf);
		}
		break;
	}
	case IOCTL_VSNAP_VDIFF_GET_RANGE:
	{
		VSNAP_VDIFF_GET_RANGE_OUTPUT_BUFFER buf;

		if (copy_from_user(&buf, param, sizeof(buf))) {
			retval = -EFAULT;
		} else {
			stcvsm_notice(LOG_CTL, "IOCTL_VSNAP_VDIFF_GET_RANGE: (%d,%d)\n", buf.major, buf.minor);
			retval = vdiff_ioctl_on_snap(MKDEV(buf.major, buf.minor), cmd, &buf);
			if (retval == 0 && copy_to_user(param, &buf, sizeof(buf)))
				retval = -EFAULT;
		}
		break;
	}
	case IOCTL_VSNAP_VDIFF_NATIVE_ON:
	{
		VSNAP_IOCTL_NATIVE_SUPP_BUFF buf;

		if (copy_from_user(&buf, param, sizeof(buf))) {
			retval = -EFAULT;
		} else {
			stcvsm_notice(LOG_CTL, "IOCTL_VSNAP_VDIFF_NATIVE_ON: (%d,%d)\n", buf.src.major, buf.src.minor);
			retval = stcvsm_ioctl_vdiff(MKDEV(buf.src.major, buf.src.minor), cmd, &buf);
		}
		break;
	}
	case IOCTL_VSNAP_VDIFF_NATIVE_OFF:
	{
		VSNAP_IOCTL_NATIVE_SUPP_BUFF buf;

		if (copy_from_user(&buf, param, sizeof(buf))) {
			retval = -EFAULT;
		} else {
			stcvsm_notice(LOG_CTL, "IOCTL_VSNAP_VDIFF_NATIVE_OFF: (%d,%d)\n", buf.dst.major, buf.dst.minor);
			retval = stcvsm_ioctl_vdiff(MKDEV(buf.dst.major, buf.dst.minor), cmd, &buf);
		}
		break;
	}
	case IOCTL_SNAP_VOLUME_EX:
	{
		VSNAP_SNAP_INPUT_BUFFER_EX buf;

		if (copy_from_user(&buf, param, sizeof(buf))) {
			retval = -EFAULT;
		} else {
			stcvsm_notice(LOG_CTL, "IOCTL_SNAP_VOLUME_EX: (%d,%d) (%d,%d)\n",
					buf.major, buf.minor, buf.cow_major, buf.cow_minor);
			retval = stcvsm_ioctl_snapshot(MKDEV(buf.major, buf.minor), cmd, &buf);
		}
		break;
	}
	case IOCTL_UNSNAP_VOLUME:
	{
		MAJOR_MINOR buf;

		if (copy_from_user(&buf, param, sizeof(buf))) {
			retval = -EFAULT;
		} else {
			stcvsm_notice(LOG_CTL, "IOCTL_UNSNAP_VOLUME: (%d,%d)\n", buf.major, buf.minor);
			retval = stcvsm_ioctl_snapshot(MKDEV(buf.major, buf.minor), cmd, NULL);
		}
		break;
	}
	case IOCTL_GET_SNAPSHOT_NAME_EX:
	{
		SNAPSHOT_NAME_BUF __user *name = (SNAPSHOT_NAME_BUF __user *)param;
		int major, minor, avail;
		SNAPSHOT_NAME_BUF buf;

		if (get_user(major, &name->major) || get_user(minor, &name->minor) ||
				get_user(avail, &name->path_size)) {
			retval = -EFAULT;
		} else {
			stcvsm_notice(LOG_CTL, "IOCTL_GET_SNAPSHOT_NAME_EX: (%d,%d)\n", major, minor);
			retval = stcvsm_ioctl_snapshot(MKDEV(major, minor), cmd, &buf);
			if (retval)
				break;
			retval = put_user(buf.path_size, &name->path_size);
			if (retval)
				break;
			if (buf.path_size > avail) {
				retval = -ENOBUFS;
			} else {
				retval = copy_to_user(name->path, buf.path, buf.path_size);
				if (retval)
					retval = -EFAULT;
			}
		}
		break;
	}
#ifdef RUNTIME_LOG
	case IOCTL_SET_LOG_LEVEL:
	{
		STCVSM_LOG_LEVEL_DATA __user *log_level = (STCVSM_LOG_LEVEL_DATA __user *)param;
		int type, level;

		if (get_user(type, &log_level->type) || get_user(level, &log_level->level)) {
			retval = -EFAULT;
		} else {
			set_log_level(type, level);
			retval = 0;
		}
		break;
	}
#endif
	case IOCTL_VSNAP_SET_ADD_DEV:
	{
		int major, minor, cow_major, cow_minor;
		GUID set_guid;

		VSNAP_IOCTL_SET_DEV __user *set_dev = (VSNAP_IOCTL_SET_DEV __user *)param;

		if (get_user(major, &set_dev->major) || get_user(minor, &set_dev->minor) ||
				get_user(cow_major, &set_dev->cow_major) ||
				get_user(cow_minor, &set_dev->cow_minor) ||
				copy_from_user(set_guid, set_dev->SetGuid, 16)) {
			retval = -EFAULT;
		} else {
			stcvsm_notice(LOG_CTL, "IOCTL_VSNAP_SET_ADD_DEV: (%d,%d) (%d,%d)\n",
					major, minor, cow_major, cow_minor);
			printk_uuid(info, "Guid", set_guid);
			retval = set_add(set_guid, MKDEV(major, minor),
					MKDEV(cow_major, cow_minor));
		}
		break;
	}
	case IOCTL_VSNAP_SET_SNAP:
	{
		GUID set_guid;
		GUID __user *uset_guid = (GUID __user *)param;

		if (copy_from_user(set_guid, uset_guid, 16)) {
			retval = -EFAULT;
		} else {
			printk_uuid(info, "IOCTL_VSNAP_SET_SNAP : Guid", set_guid);
			retval = set_snap(set_guid);
			if (retval)
				stcvsm_notice(LOG_CTL, "IOCTL_VSNAP_SET_SNAP err: %ld", retval);
		}
		break;
	}
	case IOCTL_VSNAP_SET_UNSNAP:
	{
		GUID set_guid;
		GUID __user *uset_guid = (GUID __user *)param;

		if (copy_from_user(set_guid, uset_guid, 16)) {
			retval = -EFAULT;
		} else {
			printk_uuid(info, "IOCTL_VSNAP_SET_UNSNAP: Guid", set_guid);
			retval = set_unsnap(set_guid);
		}
		break;
	}
	default:
	{
		retval = -EOPNOTSUPP;
		break;
	}
	}
	return retval;
}

#ifdef CONFIG_COMPAT
static long stcvsmctl_ioctl_compat(struct file *file, uint command, ulong u)
{
	return (long)stcvsmctl_ioctl(file, command, (ulong) compat_ptr(u));
}
#else
#define stcvsmctl_ioctl_compat NULL
#endif

static const struct file_operations stcvsmctl_ops = {
	.open		= nonseekable_open,
	.unlocked_ioctl	= stcvsmctl_ioctl,
	.compat_ioctl	= stcvsmctl_ioctl_compat,
	.owner		= THIS_MODULE,
};

static struct miscdevice stcvsmctl_dev = {
	.minor		= MISC_DYNAMIC_MINOR,
	.name		= STCVSMCTL_DEVICE_NAME,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 31)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 32)
	.devnode	= STCVSMCTL_DEVICE_NAME,
#else
	.nodename	= STCVSMCTL_DEVICE_NAME,
#endif
#endif
	.fops		= &stcvsmctl_ops
};

int __init stcvsmctl_init(void)
{
	int err;

	err = misc_register(&stcvsmctl_dev);
	if (err)
		stcvsm_err(LOG_CTL, "Unable to misc_register the control device.\n");
	return err;
}

void stcvsmctl_destroy(void)
{
	misc_deregister(&stcvsmctl_dev);
}
