/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

/*
 * Utility package - tree table
 * 64bit ready
 */
#ifndef _TREE_H
#define _TREE_H

#include "stcvsm_os.h"
#include "look_aside_list.h"

/* Types */

/* Table kind enumeration */
enum tree_kind {

	/*
	 * AVL tree
	 * Guaranteed to be optimally balanced
	 * A maximum of 2 turns per insert
	 * Delete can need lots of turns
	 */
	enum_tree_avl_kind,

#ifdef ENABLE_RED_BLACK
	/*
	 * Red-black tree
	 * Guaranteed to be not worse then 2 times suboptimally balanced
	 * A maximum of 2 turns per insert
	 * A maximum of 3 turns per delete
	 */
	enum_tree_red_black_kind
#endif
};

/*
 * Allocate routine for a tree table
 * Mandatory
 * Parameter is allocator passed to tree_inialize_xxxx
 * Must return NULL if failed, or the pool (8byte) aligned block address
 */
typedef void *(*tree_allocate_fn)(void *);
/*
 * Constructor routine for a tree table item
 * Optional
 * Called when a new item is added
 * First parameter is a tree table
 * Second parameter is a a data area of the item being inserted (to be filled)
 * Third parameter is a Data passed to tree_insert (construction parameters)
 */
typedef void (*tree_construct_fn)(void *, void *, void *);
/*
 * Destructor routine for a tree table item
 * Optional
 * Called before the free routine to clean the item up
 * First parameter is a tree table
 * Second parameter is a a data area of the item being freed
 */
typedef void (*tree_destroy_fn)(void *, void *);
/*
 * Free routine for a tree table
 * Optional (arena has no it)
 * Called to free the item's memory
 * First parameter is allocator passed to tree_initialize
 * Second parameter is the item to free
 * Note that this matches ExFreeToxxxLookasideList
 */
typedef void (*tree_free_fn)(void *, void *);
/*
 * Cleanup routine for a tree table
 * Optional
 * Called to clean up an allocator
 * Parameter is allocator passed to tree_initialize
 */
typedef void (*tree_cleanup_fn)(void *);
/*
 * Compare routine for a tree table
 * Mandatory usually, optional only if FAST_UT_TREE is defined, in this case, absense of
 *  this routine means - data areas start with u32 which are considered to be indices
 * First parameter is a tree table structure
 * Second and third are data areas of 2 items
 * Returns:
 *  - negative if <
 *  - 0 if ==
 *  - positive if >
 */
typedef s32 (*tree_compare_fn)(void *, void *, void *);

/* Structures */

struct splay_links {
	struct splay_links *parent;
	struct splay_links *left_child;
	struct splay_links *right_child;
};

#define Parent(links) ((links)->parent)
#define LeftChild(links) ((links)->left_child)
#define RightChild(links) ((links)->right_child)
#define IsRightChild(links) (links == (links)->parent->right_child)
#define IsLeftChild(links) (links == (links)->parent->left_child)
#define IsRoot(links) ((links)->parent == links)


static inline void initialize_splay_links(struct splay_links *links)
{
	links->parent = links;
	links->left_child = NULL;
	links->right_child = NULL;
}

/* Successor of the tree item */
static inline struct splay_links *tree_real_successor(struct splay_links *links)
{
	struct splay_links *item;

	/* Take the right child */
	item = RightChild(links);
	if (item != NULL) {
		/*
		 * There is a subtree successor
		 * Go left
		 */
		while (LeftChild(item) != NULL)
			item = LeftChild(item);
		/* Done */
		return item;
	}
	/*
	 * There is no subtree successor
	 * Assume that we are a subtree predecessor of some item
	 * Go up while we are a right child
	 */
	item = links;
	while (IsRightChild(item))
		item = Parent(item);
	/* item is not a right child */
	if (IsLeftChild(item))
		/* Found */
		return Parent(item);
	/*
	 * item is not a left and not a right child
	 * item is root, links was the rightmost item
	 * No successor
	 */
	return NULL;
}

/* Predecessor of the tree item */
static inline struct splay_links *tree_real_predecessor(struct splay_links *links)
{
	struct splay_links *item;

	/* Take the left child */
	item = LeftChild(links);
	if (item != NULL) {
		/*
		 * There is a subtree predecessor
		 * Go right
		 */
		while (RightChild(item) != NULL)
			item = RightChild(item);
		/* Done */
		return item;
	}
	/*
	 * There is no subtree predecessor
	 * Assume that we are a subtree successor of some item
	 * Go up while we are a left child
	 */
	item = links;
	while (IsLeftChild(item))
		item = Parent(item);
	/* item is not a left child */
	if (IsRightChild(item))
		/* Found */
		return Parent(item);
	/*
	 * item is not a left and not a right child
	 * item is root, links was the leftmost item
	 * No predecessor
	 */
	return NULL;
}

/* Subtree successor of the tree item */
static inline struct splay_links *tree_subtree_successor(struct splay_links *links)
{
	struct splay_links *item;

	/* Take the right child */
	item = RightChild(links);
	if (item == NULL)
		/* No subtree successor */
		return NULL;
	/* Go left */
	while (LeftChild(item) != NULL)
		item = LeftChild(item);
	/* Done */
	return item;
}


/* Tree table item */
struct tree_table_item {
	/* Tree links */
	struct splay_links links;
	/*
	 *  0 is black, 1 is red for red-black tree
	 * Balance factor for AVL tree
	 */
	s32 value;
	/* Data follows */
};

/* Internal rebalancing engine handler routines - different for Splay/RB/AVL trees */
typedef void (*tree_handler_fn)(struct tree_table_item *, struct splay_links **);
/* Called while the deleted item is already detached from the tree */
typedef void (*tree_delete_handler_fn)(struct tree_table_item *, struct tree_table_item *,
					struct tree_table_item *, u32, u32,
					struct splay_links **);

/* Tree table */
struct tree_table {
	/* allocator */
	void *allocator;
	/* item allocator routine */
	tree_allocate_fn allocate_func;
	/* item constructor routine */
	tree_construct_fn constructor_func;
	/* item destructor routine */
	tree_destroy_fn destroy_func;
	/* item deallocator routine */
	tree_free_fn free_func;
	/* allocator cleanup routine */
	tree_cleanup_fn cleanup_func;
	/* Kind-specific insert handler */
	tree_handler_fn insert_handler_func;
	/* Kind-specific delete handler */
	tree_delete_handler_fn delete_handler_func;
	/* Kind-specific lookup handler */
	tree_handler_fn lookup_handler_func;
	/* Comparison */
	tree_compare_fn compare_func;
	/* Data area size to be copied to the new item */
	u32 data_area_size;
	/* item count */
	u32 item_count;
	/* Root item */
	struct splay_links *root;
};

/* Macros */

/* Full item size used by the caller to set up an allocator */
#define TreeFullItemSize(data_area_size)    ((data_area_size) + sizeof(struct tree_table_item))
/* item count in the tree */
#define TreeItemCount(Tree)               ((Tree)->item_count)

/* Functions */

/* Initializes a generic tree table with caller-specified pointers and allocator semantics */
void tree_initialize_generic(struct tree_table *tree, enum tree_kind tree_kind, u32 data_area_size,
			     tree_compare_fn compare_func,
			     tree_allocate_fn allocate_func,
			     tree_construct_fn constructor_func,
			     tree_destroy_fn destroy_func,
			     tree_free_fn free_func,
			     tree_cleanup_fn cleanup_func,
			     void *allocator);
/*
 * Initializes a tree table with a nonpaged lookaside allocator provided by the caller
 * Lookaside is initialized too
 */
void tree_initialize_npage(struct tree_table *Tree, enum tree_kind Kind, u32 data_area_size,
			    tree_compare_fn compare_func,
			    tree_construct_fn constructor_func,
			    tree_destroy_fn destroy_func,
			    struct lookaside_list *lookaside,
			    u32 pool_tag);
/* Same as tree_initialize_npage but using already initialized lookaside */
void tree_initialize_npage_with_lookaside(struct tree_table *tree, enum tree_kind tree_kind,
					  u32 data_area_size,
					  tree_compare_fn compare_func,
					  tree_construct_fn constructor_func,
					  tree_destroy_fn destroy_func,
					  struct lookaside_list *lookaside);

/*
 * Copies the tree table from one place to another
 * Src is empty after the copy, and shares the common allocator with Dst
 *  (must only be cleaned up once)
 */
void tree_copy(struct tree_table *dst, struct tree_table *src);
/*
 * Cleans up a tree table, deleting all items and cleaning up an allocator
 * Both allocator and a tree table can be reused then
 */
void tree_cleanup(struct tree_table *tree);
/*
 * Deletes all items from the tree table
 * The table is still initialized and empty after this
 */
void tree_delete_all(struct tree_table *tree);
/*
 * Inserts an item described by Data to the tree
 * Duplicate is ignored
 * Returns the item inserted or duplicate found or NULL on allocation failure
 * *NewElement is set to true if the item was inserted or to false if duplicate was found
 */
void *tree_insert(struct tree_table *tree, void *data, bool *new_element);
/*
 * Allocates the new item, but does not attach it to the tree
 * The item must later be either attached to the tree by tree_attach,
 *  or deleted by tree_prefree
 * Returns NULL on allocation error
 * The item is not filled at all
 */
void *tree_prealloc(struct tree_table *tree);
/* Undo for tree_prealloc, in a case the item will never be attached to the tree */
void tree_prefree(struct tree_table *tree, void *item);
/*
 * Attaches the item allocated by tree_prealloc to the tree
 * Data is the item's construction data
 * Same as tree_insert but uses preallocated item and thus cannot fail
 * Returns the item inserted or duplicate found
 * If duplicate was found - then returns with *NewElement set to false
 *  and calls tree_prefree(item) internally
 * Otherwise, really attaches the item to the tree and returns with *NewElement set to true
 */
void *tree_attach(struct tree_table *tree,
		   void *preallocated_item,
		   void *data,
		   bool *new_element);
/*
 * Deletes an item described by Data
 * No-op if no such item
 * Returns true if the item was really deleted, false if it was not found
 */
bool tree_delete(struct tree_table *tree, void *data);
/*
 * Deletes an item without the tree scan before deletion
 * This function requires item to be the in-tree item returned by tree_insert, tree_lookup
 *  or tree_enumerate
 */
void tree_delete_existing(struct tree_table *tree, void *item);
/*
 * Detaches the item from the tree but does not delete it
 * Useful if the destroy routine cannot be called under the lock which guards the tree
 *  (for instance, if the tree objects reference one another)
 * This function requires item to be the in-tree item returned by tree_insert, tree_lookup
 *  or tree_enumerate
 */
void tree_detach(struct tree_table *tree, void *item);
/*
 * Deletes the item which was previously detached from the tree by tree_detach
 * Does not change the tree structure and thus can be called without holding the lock
 *  which protectes the tree
 */
void tree_delete_dettached(struct tree_table *tree, void *item);
/*
 * Looks up for an item described by Data
 * Returns an item if found or NULL if not found
 */
void *tree_lookup(struct tree_table *tree, void *data);
/*
 * Looks up for an item described by Data or the nearest to it in the comparison order
 * Always returns some item except the case if the table is empty
 * *comparison_result is set to the comparison result between Data and returned item
 *  < 0 - ReturnedItem < Data
 *  == 0 - exact hit as in tree_lookup
 *  > 0 - ReturnedItem > Data
 */
void *tree_lookup_nearest(struct tree_table *tree, void *Data, s32 *comparison_result);
/*
 * Returns the next item in the comparison order or NULL if the given existing tree item
 *  is last
 */
void *tree_next(void *existing_tree_item);
/*
 * Returns the previous item in the comparison order or NULL if the given existing tree item
 *  is first
 */
void *tree_prev(void *ExistingTreeItem);
/* Returns the first tree item in the comparison order or NULL if the tree is empty */
void *tree_first(struct tree_table *tree);
/* Returns the last tree item in the comparison order or NULL if the tree is empty */
void *tree_last(struct tree_table *tree);
/*
 * Tree enumerator
 * Returns items in the comparison order
 * Does not do any balancing reorders
 * Returns the next item if found or NULL if no more items
 * *restart_key must be set to NULL to get the first item, and must be preserved for future
 *  calls
 */
void *tree_enumerate(struct tree_table *tree, void **restart_key);

#endif /* _TREE_H */
